#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.19 03:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot, Signal

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    output = Signal(object, str)
    html_output = Signal(object)
    ipython_output = Signal(object)
    build_finished = Signal()
    clear = Signal()


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        responses = []
        for item in self.cmds:
            if self.is_stopped():
                break
            try:
                response = None
                if (item["cmd"] in self.plugin.allowed_cmds
                        and (self.plugin.has_cmd(item["cmd"]) or 'force' in item)):

                    if item["cmd"] == "sys_exec":
                        response = self.cmd_sys_exec(item)

                    if response:
                        responses.append(response)

            except Exception as e:
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

    def cmd_sys_exec(self, item: dict) -> dict:
        """
        Execute system command

        :param item: command item
        :return: response item
        """
        request = self.from_request(item)
        try:
            if not self.plugin.runner.is_sandbox():
                result = self.plugin.runner.sys_exec_host(
                    ctx=self.ctx,
                    item=item,
                    request=request,
                )
            else:
                result = self.plugin.runner.sys_exec_sandbox(
                    ctx=self.ctx,
                    item=item,
                    request=request,
                )
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def prepare_extra(self, item: dict, result: dict) -> dict:
        """
        Prepare extra data for response

        :param item: command item
        :param result: response data
        :return: extra data
        """
        cmd = item["cmd"]
        extra = {
            'plugin': "cmd_system",
            'cmd': cmd,
            'code': {}
        }
        lang = "python"
        if cmd in ["sys_exec"]:
            lang = "bash"
        if "params" in item and "code" in item["params"]:
            extra["code"]["input"] = {}
            extra["code"]["input"]["lang"] = lang
            extra["code"]["input"]["content"] = str(item["params"]["code"])
        if "result" in result:
            extra["code"]["output"] = {}
            extra["code"]["output"]["lang"] = lang
            extra["code"]["output"]["content"] = str(result["result"])
        if "context" in result:
            extra["context"] = str(result["context"])
        return extra





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.18 23:00:00                  #
# ================================================== #

import os

from PySide6.QtCore import QTimer

from pygpt_net.core.events import AppEvent
from pygpt_net.core.tabs.tab import Tab
from pygpt_net.utils import trans


class Simple:

    MIN_FRAMES = 25  # minimum frames to start transcription

    def __init__(self, plugin=None):
        """
        Simple audio input handler

        :param plugin: plugin instance
        """
        self.plugin = plugin
        self.is_recording = False
        self.timer = None

    def toggle_recording(self):
        """Toggle recording"""
        if self.is_recording:
            self.stop_recording()
        else:
            self.start_recording()

    def switch_btn_stop(self):
        """Switch button to stop"""
        self.plugin.window.ui.plugin_addon['audio.input.btn'].btn_toggle.setText(trans('audio.speak.btn.stop'))
        self.plugin.window.ui.plugin_addon['audio.input.btn'].btn_toggle.setToolTip(
            trans('audio.speak.btn.stop.tooltip'))

    def switch_btn_start(self):
        """Switch button to start"""
        self.plugin.window.ui.plugin_addon['audio.input.btn'].btn_toggle.setText(trans('audio.speak.btn'))
        self.plugin.window.ui.plugin_addon['audio.input.btn'].btn_toggle.setToolTip(trans('audio.speak.btn.tooltip'))

    def stop_timeout(self):
        """Stop timeout"""
        self.stop_recording(timeout=True)

    def start_recording(self, force: bool = False):
        """
        Start recording

        :param force: True to force recording
        """
        # display snap warning if not displayed yet
        if (not self.plugin.window.core.config.get("audio.input.snap", False)
                or not self.plugin.window.core.config.has("audio.input.snap")):
            if self.plugin.window.core.platforms.is_snap():
                self.plugin.window.ui.dialogs.open(
                    'snap_audio_input',
                    width=400,
                    height=200
                )
                self.plugin.window.core.config.set("audio.input.snap", True)
                self.plugin.window.core.config.save()
                return

        # enable continuous mode if notepad tab is active
        self.plugin.window.core.audio.capture.stop_callback = self.on_stop
        continuous_enabled = self.plugin.window.core.config.get('audio.input.continuous', False)
        if continuous_enabled and self.plugin.window.controller.ui.tabs.get_current_type() == Tab.TAB_NOTEPAD:
            self.plugin.window.core.audio.capture.loop = True  # set loop
        else:
            self.plugin.window.core.audio.capture.loop = False

        try:
            # stop audio output if playing
            self.plugin.window.controller.audio.stop_output()

            # set audio volume bar
            self.plugin.window.core.audio.capture.set_bar(
                self.plugin.window.ui.plugin_addon['audio.input.btn'].bar
            )

            # start timeout timer to prevent infinite recording
            # disable in continuous mode
            timeout = int(self.plugin.window.core.config.get('audio.input.timeout', 120) or 0) # get timeout
            timeout_continuous = self.plugin.window.core.config.get('audio.input.timeout.continuous', False) # enable continuous timeout
            if timeout > 0:
                if self.timer is None and (not continuous_enabled or timeout_continuous):
                    self.timer = QTimer()
                    self.timer.timeout.connect(self.stop_timeout)
                    self.timer.start(timeout * 1000)

            if not force:
                if not self.plugin.window.core.audio.capture.check_audio_input():
                    raise Exception("Audio input not working.")
                    # IMPORTANT!!!!
                    # Stop here if audio input not working!
                    # This prevents the app from freezing when audio input is not working!

            self.is_recording = True
            self.switch_btn_stop()
            self.plugin.window.core.audio.capture.start()  # start recording if audio is OK
            self.plugin.window.update_status(trans('audio.speak.now'))
            self.plugin.window.dispatch(AppEvent(AppEvent.INPUT_VOICE_LISTEN_STARTED))  # app event
        except Exception as e:
            self.is_recording = False
            self.plugin.window.core.debug.log(e)
            self.plugin.window.ui.dialogs.alert(e)
            if self.plugin.window.core.platforms.is_snap():
                self.plugin.window.ui.dialogs.open(
                    'snap_audio_input',
                    width=400,
                    height=200
                )
            self.switch_btn_start()  # switch button to start

    def stop_recording(self, timeout: bool = False):
        """
        Stop recording

        :param timeout: True if stopped due to timeout
        """
        self.plugin.window.core.audio.capture.reset_audio_level()
        self.is_recording = False
        self.plugin.window.dispatch(AppEvent(AppEvent.INPUT_VOICE_LISTEN_STOPPED))  # app event
        if self.timer:
            self.timer.stop()
            self.timer = None
        self.switch_btn_start()  # switch button to start
        path = os.path.join(self.plugin.window.core.config.path, self.plugin.input_file)
        self.plugin.window.core.audio.capture.set_path(path)

        if self.plugin.window.core.audio.capture.has_source():
            self.plugin.window.core.audio.capture.stop()  # stop recording
            # abort if timeout
            if timeout:
                self.plugin.window.update_status("Aborted.".format(timeout))
                return

            if self.plugin.window.core.audio.capture.has_frames():
                frames = self.plugin.window.core.audio.capture.get_frames()
                if len(frames) < self.MIN_FRAMES:
                    self.plugin.window.update_status(trans("status.audio.too_short"))
                    self.plugin.window.dispatch(AppEvent(AppEvent.VOICE_CONTROL_STOPPED))  # app event
                    return

                self.plugin.handle_thread(True)  # handle transcription in simple mode
        else:
            self.plugin.window.update_status("")


    def on_stop(self):
        """Handle auto-transcribe"""
        path = os.path.join(self.plugin.window.core.config.path, self.plugin.input_file)
        self.plugin.window.core.audio.capture.set_path(path)
        self.plugin.window.core.audio.capture.stop()
        self.plugin.window.core.audio.capture.start()
        self.plugin.handle_thread(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import uuid
from typing import Dict

from packaging.version import Version

from pygpt_net.item.notepad import NotepadItem
from pygpt_net.provider.core.notepad.base import BaseProvider
from .patch import Patch
from .storage import Storage


class DbSqliteProvider(BaseProvider):
    def __init__(self, window=None):
        super(DbSqliteProvider, self).__init__(window)
        self.window = window
        self.patcher = Patch(window, self)
        self.storage = Storage(window)
        self.id = "db_sqlite"
        self.type = "notepad"

    def attach(self, window):
        self.window = window
        self.storage.attach(window)

    def patch(self, version: Version) -> bool:
        """
        Patch versions

        :param version: current app version
        :return: True if migrated
        """
        return self.patcher.execute(version)

    def create_id(self) -> str:
        """
        Create unique uuid

        :return: uuid
        """
        return str(uuid.uuid4())

    def create(self, notepad: NotepadItem) -> int:
        """
        Create new and return its ID

        :param notepad: NotepadItem
        :return: notepad ID
        """
        if notepad.id is None or notepad.id == "":
            notepad.id = self.storage.insert(notepad)
        return notepad.id

    def load_all(self) -> Dict[int, NotepadItem]:
        """
        Load notepads from DB

        :return: notepads
        """
        return self.storage.get_all()

    def load(self, idx: int) -> NotepadItem:
        """
        Load notepad from DB

        :param idx: notepad IDx
        :return: notepad
        """
        return self.storage.get_by_idx(idx)

    def save(self, notepad: NotepadItem):
        """
        Save notepad to DB

        :param notepad: NotepadItem
        """
        try:
            self.storage.save(notepad)
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error while saving notepad: {}".format(str(e)))

    def save_all(self, items: Dict[int, NotepadItem]):
        """
        Save all notepads to DB

        :param items: dict of NotepadItem objects
        """
        try:
            for idx in items:
                notepad = items[idx]
                self.storage.save(notepad)
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error while saving notepad: {}".format(str(e)))

    def truncate(self) -> bool:
        """
        Truncate all notepads

        :return: True if truncated
        :rtype: bool
        """
        return self.storage.truncate_all()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 01:00:00                  #
# ================================================== #

import json
from typing import List, Dict, Any

from llama_index.core.chat_engine.types import AgentChatResponse
from llama_index.core.tools import BaseTool, FunctionTool, QueryEngineTool, ToolMetadata

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem


class Tools:
    def __init__(self, window=None):
        """
        Agent tools

        :param window: Window instance
        """
        self.window = window
        self.cmd_blacklist = []
        self.verbose = False

    def prepare(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False
    ) -> List[BaseTool]:
        """
        Prepare tools for agent

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        :return: list of tools
        """
        self.verbose = verbose
        tools = []

        # add functions from plugins
        plugin_functions = self.get_plugin_functions(context.ctx, verbose=verbose)
        tools.extend(plugin_functions)

        # add query engine tool if idx is provided
        idx = extra.get("agent_idx", None)
        if idx is not None and idx != "_":
            llm, embed_model = self.window.core.idx.llm.get_service_context(model=context.model)
            index = self.window.core.idx.storage.get(idx, llm, embed_model)  # get index
            if index is not None:
                query_engine = index.as_query_engine(similarity_top_k=3)
                query_engine_tools = [
                    QueryEngineTool(
                        query_engine=query_engine,
                        metadata=ToolMetadata(
                            name="query_engine",
                            description=(
                                "Provides additional context and access to the indexed documents."
                            ),
                        ),
                    ),
                ]
                tools.extend(query_engine_tools)
        return tools

    def get_plugin_functions(
            self,
            ctx: CtxItem,
            verbose: bool = False
    ) -> list:
        """
        Parse plugin functions

        :param ctx: CtxItem
        :param verbose: verbose mode
        :return: List of functions
        """
        tools = []
        functions = self.window.core.command.get_functions()
        for func in functions:
            try:
                name = func['name']
                if name in self.cmd_blacklist:
                    continue  # skip blacklisted commands

                description = func['desc']
                schema = json.loads(func['params'])  # from JSON to dict

                def make_func(name):
                    def func(**kwargs):
                        self.log("[Plugin] Tool call: " + name + " " + str(kwargs))
                        cmd = {
                            "cmd": name,
                            "params": kwargs,
                        }
                        return self.window.controller.plugins.apply_cmds_all(
                            ctx,  # current ctx
                            [cmd],  # commands
                        )

                    return func

                func = make_func(name)
                metadata = PluginToolMetadata(
                    name=name,
                    description=description,
                )
                metadata.schema = schema
                tool = FunctionTool(
                    fn=func,
                    metadata=metadata,
                )
                tools.append(tool)
            except Exception as e:
                print(e)
        return tools

    def export_sources(
            self,
            response: AgentChatResponse
    ) -> List[dict]:
        """
        Export sources from response

        :param response: response
        :return: list of sources
        """
        data = []
        for output in response.sources:
            item = {}
            item["ToolOutput"] = {
                "content": str(output.content),
                "tool_name": str(output.tool_name),
                "raw_input": str(output.raw_input),
                "raw_output": str(output.raw_output),
            }
            data.append(item)
        return data

    def log(self, msg: str):
        """
        Log message

        :param msg: message
        """
        if self.verbose:
            print(msg)
            self.window.core.debug.add(msg)

class PluginToolMetadata(ToolMetadata):
    def __init__(self, name: str, description: str):
        super().__init__(name=name, description=description)
        self.schema = None

    def get_parameters_dict(self) -> Dict[str, Any]:
        """
        Get parameters dictionary

        :return: parameters
        """
        parameters = {
            k: v
            for k, v in self.schema.items()
            if k in ["type", "properties", "required", "definitions"]
        }
        return parameters


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 02:00:00                  #
# ================================================== #

import os

from PySide6 import QtCore
from PySide6.QtGui import QStandardItemModel, Qt, QIcon
from PySide6.QtWidgets import QVBoxLayout, QPushButton, QHBoxLayout, QWidget, QRadioButton

from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.lists.attachment_ctx import AttachmentCtxList
from pygpt_net.utils import trans

import pygpt_net.icons_rc

class AttachmentsCtx:
    def __init__(self, window=None):
        """
        Attachments for CTX UI

        :param window: Window instance
        """
        self.window = window
        self.id = 'attachments_ctx'

    def setup(self) -> QVBoxLayout:
        """
        Setup list

        :return: QVBoxLayout
        """
        self.setup_attachments()
        empty_widget = QWidget()

        self.window.ui.nodes['tip.input.attachments.ctx'] = HelpLabel(trans('tip.input.attachments.ctx'),
                                                                           self.window)

        self.window.ui.nodes['input.attachments.ctx.mode.label'] = HelpLabel(trans("attachments.ctx.label"))
        self.window.ui.nodes['input.attachments.ctx.mode.full'] = QRadioButton(trans("attachments.ctx.mode.full"))
        self.window.ui.nodes['input.attachments.ctx.mode.full'].clicked.connect(
            lambda: self.window.controller.chat.attachment.switch_mode(
                self.window.controller.chat.attachment.MODE_FULL_CONTEXT
            ))
        self.window.ui.nodes['input.attachments.ctx.mode.query'] = QRadioButton(trans("attachments.ctx.mode.query"))
        self.window.ui.nodes['input.attachments.ctx.mode.query'].clicked.connect(
            lambda: self.window.controller.chat.attachment.switch_mode(
                self.window.controller.chat.attachment.MODE_QUERY_CONTEXT
            ))
        self.window.ui.nodes['input.attachments.ctx.mode.query_summary'] = QRadioButton(trans("attachments.ctx.mode.summary"))
        self.window.ui.nodes['input.attachments.ctx.mode.query_summary'].clicked.connect(
            lambda: self.window.controller.chat.attachment.switch_mode(
                self.window.controller.chat.attachment.MODE_QUERY_CONTEXT_SUMMARY
            ))
        self.window.ui.nodes['input.attachments.ctx.mode.off'] = QRadioButton(trans("attachments.ctx.mode.off"))
        self.window.ui.nodes['input.attachments.ctx.mode.off'].clicked.connect(
            lambda: self.window.controller.chat.attachment.switch_mode(
                self.window.controller.chat.attachment.MODE_DISABLED
            ))

        # buttons layout
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self.window.ui.nodes['attachments_ctx.btn.clear'])
        buttons_layout.addWidget(empty_widget)
        # buttons_layout.addStretch()
        buttons_layout.addWidget(self.window.ui.nodes['input.attachments.ctx.mode.label'])
        buttons_layout.addWidget(self.window.ui.nodes['input.attachments.ctx.mode.full'])
        buttons_layout.addWidget(self.window.ui.nodes['input.attachments.ctx.mode.query'])
        buttons_layout.addWidget(self.window.ui.nodes['input.attachments.ctx.mode.query_summary'])
        buttons_layout.addWidget(self.window.ui.nodes['input.attachments.ctx.mode.off'])
        buttons_layout.addStretch()

        # layout
        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes['tip.input.attachments.ctx'])
        layout.addWidget(self.window.ui.nodes['attachments_ctx'])
        layout.addLayout(buttons_layout)

        return layout

    def setup_attachments(self):
        """Setup attachments uploaded list"""
        # attachments
        self.window.ui.nodes[self.id] = AttachmentCtxList(self.window)

        # buttons
        self.window.ui.nodes['attachments_ctx.btn.clear'] = QPushButton(QIcon(":/icons/close.svg"), trans('attachments_uploaded.btn.clear'))
        self.window.ui.nodes['attachments_ctx.btn.clear'].clicked.connect(
            lambda: self.window.controller.chat.attachment.clear()
        )

        self.window.ui.models[self.id] = self.create_model(self.window)
        self.window.ui.nodes[self.id].setModel(self.window.ui.models[self.id])

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model

        :param parent: parent widget
        :return: QStandardItemModel
        """
        model = QStandardItemModel(0, 5, parent)
        model.setHeaderData(0, Qt.Horizontal, trans('attachments.header.name'))
        model.setHeaderData(1, Qt.Horizontal, trans('attachments.header.path'))
        model.setHeaderData(2, Qt.Horizontal, trans('attachments.header.size'))
        model.setHeaderData(3, Qt.Horizontal, trans('attachments.header.length'))
        model.setHeaderData(4, Qt.Horizontal, trans('attachments.header.idx'))
        return model

    def update(self, data):
        """
        Update list

        :param data: Data to update
        """
        self.window.ui.models[self.id].removeRows(0, self.window.ui.models[self.id].rowCount())
        i = 0
        for item in data:
            indexed = False
            name = "No name"
            if 'name' in item:
                name = item['name']
            size = "-"
            path = "No path"
            if 'path' in item:
                path = item['path']
            uuid = ""
            if 'uuid' in item:
                uuid = item['uuid']
            length = "-"
            if 'length' in item:
                length = str(item['length'])
            if 'tokens' in item:
                length += ' / ~' + str(item['tokens'])
            if 'indexed' in item and item['indexed']:
                indexed = True

            idx_str = ""
            if indexed:
                idx_str = trans("attachments.ctx.indexed")

            if os.path.exists(path):
                size = self.window.core.filesystem.sizeof_fmt(os.path.getsize(path))
            elif 'size' in item:
                size = self.window.core.filesystem.sizeof_fmt(item['size'])

            self.window.ui.models[self.id].insertRow(i)
            index = self.window.ui.models[self.id].index(i, 0)
            self.window.ui.models[self.id].setData(index, "uuid: " + str(uuid), QtCore.Qt.ToolTipRole)
            self.window.ui.models[self.id].setData(self.window.ui.models[self.id].index(i, 0), name)
            self.window.ui.models[self.id].setData(self.window.ui.models[self.id].index(i, 1), path)
            self.window.ui.models[self.id].setData(self.window.ui.models[self.id].index(i, 2), size)
            self.window.ui.models[self.id].setData(self.window.ui.models[self.id].index(i, 3), length)
            self.window.ui.models[self.id].setData(self.window.ui.models[self.id].index(i, 4), idx_str)
            i += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

import datetime
import time
from typing import Optional


class External:
    def __init__(self, window=None, provider=None):
        """
        External data core

        :param window: Window instance
        :param provider: provider name
        """
        self.window = window
        self.provider = provider

    def append(
            self,
            store_id: str,
            idx: str,
            content: str,
            type: str,
            doc_id: str
    ) -> int:
        """
        Append file to index

        :param store_id: store id
        :param idx: index name
        :param content: content
        :param type: content type
        :param doc_id: document id
        :return: ID of appended external content
        """
        data = {
            "content": content,
            "type": type,
            "indexed_ts": datetime.datetime.now().timestamp(),
            "id": doc_id,
        }
        return self.provider.append_external(
            store_id=store_id,
            idx=idx,
            data=data,
        )

    def get_doc_id(
            self,
            store_id: str,
            idx: str,
            content: str,
            type: str
    ) -> str:
        """
        Get indexed document id by content

        :param store_id: store id
        :param idx: index name
        :param content: content
        :param type: content type
        :return: document id
        """
        return self.provider.get_external_doc_id(
            store_id=store_id,
            idx=idx,
            content=content,
            type=type,
        )

    def exists(
            self,
            store_id: str,
            idx: str,
            content: str,
            type: str
    ) -> bool:
        """
        Check if content is indexed

        :param store_id: store id
        :param idx: index name
        :param content: content
        :param type: content type
        :return: True if external content is indexed
        """
        return self.provider.is_external_indexed(
            store_id=store_id,
            idx=idx,
            content=content,
            type=type,
        )

    def update(
            self,
            content: str,
            type: str,
            doc_id: str,
            ts: int
    ) -> bool:
        """
        Update timestamp of indexed external content

        :param content: content
        :param type: content type
        :param doc_id: document ID
        :param ts: timestamp
        :return: True if content was updated
        """
        return self.provider.update_external(
            content=content,
            type=type,
            doc_id=doc_id,
            ts=ts,
        )

    def remove(
            self,
            store_id: str,
            idx: str,
            doc_id: str
    ):
        """
        Remove document from index

        :param store_id: store id
        :param idx: index name
        :param doc_id: document id
        """
        self.provider.remove_external(
            store_id=store_id,
            idx=idx,
            doc_id=doc_id,
        )

    def truncate(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ):
        """
        Truncate external data from index

        :param store_id: store id
        :param idx: index name
        """
        self.provider.truncate_external(
            store_id=store_id,
            idx=idx,
        )

    def set_indexed(
            self,
            content: str,
            type: str,
            idx: str,
            doc_id: str
    ) -> bool:
        """
        Set external content as indexed

        :param content: content, e.g. url address
        :param type: content type, e.g. url
        :param idx: index name
        :param doc_id: document ID
        :return: True if updated
        """
        ts = int(time.time())
        store = self.window.core.idx.get_current_store()

        # add to index db
        if not self.exists(store, idx, content, type):
            self.append(
                store_id=store,  # current store
                idx=idx,  # index name
                content=content,  # content (e.g. url)
                type=type,  # content type
                doc_id=doc_id,  # document id
            )
        else:
            # update document id in index db if already indexed
            self.update(
                content=content,  # content (e.g. url)
                type=type,  # content type
                doc_id=doc_id,  # document id
                ts=ts,  # timestamp
            )
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_AUDIO,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_CHAT,
    MODE_RESEARCH,
)
from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.item.ctx import CtxItem
from pygpt_net.core.events import Event

from .config import Config
from .worker import Worker

class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "openai_vision"
        self.name = "GPT-4 Vision (inline)"
        self.type = [
            "vision",
            "cmd.inline",
        ]
        self.description = "Integrates GPT-4 Vision abilities with any chat mode"
        self.prefix = "Vision"
        self.order = 100
        self.use_locale = True
        self.prompt = ""
        self.allowed_urls_ext = [
            ".jpg",
            ".png",
            ".jpeg",
            ".gif",
            ".webp",
        ]
        self.allowed_cmds = [
            "camera_capture",
            "make_screenshot",
            "analyze_image_attachment",
            "analyze_screenshot",
            "analyze_camera_capture",
        ]
        self.disabled_mode_switch = [
            MODE_CHAT,
            MODE_VISION,
            MODE_AGENT,
            MODE_AGENT_LLAMA,
            MODE_LLAMA_INDEX,
            MODE_LANGCHAIN,
            MODE_AUDIO,
            MODE_RESEARCH,
        ]
        self.worker = None
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def is_allowed(self, mode: str) -> bool:
        """
        Check if plugin is allowed in given mode

        :param mode: mode name
        :return: True if allowed, False otherwise
        """
        return mode in self.window.controller.chat.vision.allowed_modes

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.MODE_BEFORE:
            if self.is_allowed(data['value']):
                data['value'] = self.on_mode_before(
                    ctx,
                    mode=data['value'],
                )  # mode change

        elif name == Event.MODEL_BEFORE:
            if "mode" in data and data["mode"] == MODE_VISION:
                key = self.get_option_value("model")
                if self.window.core.models.has(key):
                    data['model'] = self.window.core.models.get(key)

        elif name == Event.PRE_PROMPT:
            if self.is_allowed(data['mode']):
                data['value'] = self.on_pre_prompt(data['value'])

        elif name == Event.INPUT_BEFORE:
            self.prompt = str(data['value'])

        elif name == Event.SYSTEM_PROMPT:
            if self.is_allowed(data['mode']):
                data['value'] = self.on_system_prompt(data['value'])

        elif name == Event.UI_ATTACHMENTS:
            mode = data["mode"]
            if mode in [MODE_AGENT, MODE_AGENT_LLAMA] and not self.window.core.config.get("cmd"):
                pass
            else:
                data['value'] = True  # allow render attachments UI elements

        elif name == Event.UI_VISION:
            if self.is_allowed(data['mode']):
                data['value'] = True  # allow render vision UI elements

        elif name in [
            Event.CTX_SELECT,
            Event.MODE_SELECT,
            Event.MODEL_SELECT,
        ]:
            self.on_toggle(False)  # always reset vision flag / disable vision mode

        elif name in [
            Event.CMD_SYNTAX,
            Event.CMD_SYNTAX_INLINE,
        ]:
            self.cmd_syntax(data)

        elif name in [
            Event.CMD_INLINE,  # inline is allowed
            Event.CMD_EXECUTE,
        ]:
            self.cmd(
                ctx,
                data['commands'],
            )

        if name == Event.AGENT_PROMPT:
            silent = False
            if 'silent' in data and data['silent']:
                silent = True
            data['value'] = self.on_agent_prompt(
                data['value'],
                silent,
            )

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        if self.has_cmd("camera_capture"):
            data['cmd'].append(self.get_cmd("camera_capture"))
        """
        if self.has_cmd("make_screenshot"):
            data['cmd'].append(self.get_cmd("make_screenshot"))
        """
        if self.has_cmd("analyze_image_attachment"):
            data['cmd'].append(self.get_cmd("analyze_image_attachment"))
        if self.has_cmd("analyze_screenshot"):
            data['cmd'].append(self.get_cmd("analyze_screenshot"))
        if self.has_cmd("analyze_camera_capture"):
            data['cmd'].append(self.get_cmd("analyze_camera_capture"))

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        needed_lock = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                if item["cmd"] == "analyze_image_attachment":
                    needed_lock = True
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        # set state: busy
        self.cmd_prepare(ctx, my_commands)

        # don't allow to clear attachments list
        if needed_lock:
            self.window.controller.attachment.lock()
        else:
            self.window.controller.attachment.unlock()

        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.cmds = my_commands
            worker.ctx = ctx
            if not self.is_async(ctx):
                worker.run()
                return
            worker.run_async()

        except Exception as e:
            self.error(e)

    def prepare_request(self, item) -> dict:
        """
        Prepare request item for result

        :param item: item with parameters
        :return: request item
        """
        return {"cmd": item["cmd"]}

    def on_toggle(self, value: bool):
        """
        Events: CTX_SELECT, MODE_SELECT, MODEL_SELECT

        :param value: vision mode state
        """
        if not value:
            self.window.controller.chat.vision.is_enabled = False

    def on_system_prompt(self, prompt: str) -> str:
        """
        Event: SYSTEM_PROMPT

        :param prompt: prompt
        :return: updated prompt
        """
        # append vision prompt only if vision is provided or enabled
        if not self.is_vision_provided():
            return prompt
        prompt = "Image attachment has been already sent.\n\n" + prompt
        return prompt

    def on_pre_prompt(self, prompt: str) -> str:
        """
        Event: PRE_PROMPT

        :param prompt: prompt
        :return: updated prompt
        """
        # append vision prompt only if vision is provided or enabled
        if not self.is_vision_provided():
            return prompt

        if self.window.core.config.get("cmd"):
            return prompt  # vision handled by command

        # replace vision prompt
        if self.get_option_value("replace_prompt"):
            return self.get_option_value("prompt")
        else:
            return prompt

    def is_vision_provided(self) -> bool:
        """
        Check if content for vision is provided (images, attachments)

        :return: True if vision is provided in this ctx
        """
        result = False
        mode = self.window.core.config.get('mode')
        attachments = self.window.core.attachments.get_all(mode)  # from global mode
        self.window.core.gpt.vision.build_content(
            str(self.prompt),
            attachments,
        )  # tmp build content, provide attachments from global mode

        built_attachments = self.window.core.gpt.vision.attachments
        built_urls = self.window.core.gpt.vision.urls

        # check for images in URLs found in prompt
        img_urls = []
        for url in built_urls:
            for ext in self.allowed_urls_ext:
                if url.lower().endswith(ext):
                    img_urls.append(url)
                    break

        if len(built_attachments) > 0 or len(img_urls) > 0:
            result = True

        return result

    def on_mode_before(self, ctx: CtxItem, mode: str) -> str:
        """
        Event: MODE_BEFORE

        :param ctx: current ctx
        :param mode: current mode
        :return: updated mode
        """
        # abort if already in vision mode or command enabled
        if mode == MODE_VISION or mode in self.disabled_mode_switch:
            return mode  # keep current mode

        # if already used in this ctx then keep vision mode
        if self.is_vision_provided():
            ctx.is_vision = True
            return MODE_VISION

        return mode  # keep current mode

    def on_agent_prompt(self, prompt: str, silent: bool = False) -> str:
        """
        Event: AGENT_PROMPT

        :param prompt: prompt
        :param silent: silent mode (no logs)
        :return: updated prompt
        """
        if not self.is_vision_provided():
            return prompt

        prompt = "Image attachment has been already sent.\n\n" + prompt
        return prompt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 01:00:00                  #
# ================================================== #

import datetime
import os.path
from typing import Optional

from llama_index.core import StorageContext
from llama_index.core.indices.base import BaseIndex
from llama_index.vector_stores.redis import RedisVectorStore

from pygpt_net.utils import parse_args
from .base import BaseStore


class RedisProvider(BaseStore):
    def __init__(self, *args, **kwargs):
        super(RedisProvider, self).__init__(*args, **kwargs)
        """
        Redis vector store provider

        :param args: args
        :param kwargs: kwargs
        """
        self.window = kwargs.get('window', None)
        self.id = "RedisVectorStore"
        self.prefix = "redis_"  # prefix for index directory
        self.indexes = {}

    def create(self, id: str):
        """
        Create empty index

        :param id: index name
        """
        path = self.get_path(id)
        if not os.path.exists(path):
            os.makedirs(path)
            self.store(id)

    def get_store(self, id: str) -> RedisVectorStore:
        """
        Get Redis vector store

        :param id: index name
        :return: RedisVectorStore instance
        """
        defaults = {
            "index_name": id,
        }
        additional_args = parse_args(
            self.window.core.config.get('llama.idx.storage.args', []),
        )
        if "index_name" in additional_args:
            del defaults["index_name"]

        return RedisVectorStore(
            **defaults,
            **additional_args
        )

    def get(
            self,
            id: str,
            llm: Optional = None,
            embed_model: Optional = None,
    ) -> BaseIndex:
        """
        Get index

        :param id: index name
        :param llm: LLM instance
        :param embed_model: Embedding model instance
        :return: index instance
        """
        if not self.exists(id):
            self.create(id)
        vector_store = self.get_store(id)
        storage_context = StorageContext.from_defaults(
            vector_store=vector_store,
        )
        self.indexes[id] = self.index_from_store(
            vector_store=vector_store,
            storage_context=storage_context,
            llm=llm,
            embed_model=embed_model,
        )
        return self.indexes[id]

    def store(
            self,
            id: str,
            index: Optional[BaseIndex] = None
    ):
        """
        Store index

        :param id: index name
        :param index: index instance
        """
        path = self.get_path(id)
        lock_file = os.path.join(path, 'store.lock')
        with open(lock_file, 'w') as f:
            f.write(id + ': ' + str(datetime.datetime.now()))
        self.get_store(id=id).persist(
            persist_path="",
        )
        self.indexes[id] = index



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 18:00:00                  #
# ================================================== #

from typing import List

class Whisper:
    def __init__(self, window=None):
        """
        Whisper core

        :param window: Window instance
        """
        self.window = window
        self.voices = [
            "alloy",
            "echo",
            "fable",
            "onyx",
            "nova",
            "shimmer",
        ]

    def get_voices(self) -> List[str]:
        """
        Get whisper voices

        :return: whisper voice name
        """
        return self.voices


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.26 23:00:00                  #
# ================================================== #

from typing import Dict, Any, List

from pygpt_net.core.types import (
    MODE_AGENT_LLAMA,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
)
from pygpt_net.utils import trans


class Placeholder:
    def __init__(self, window=None):
        """
        Configuration placeholder options controller

        :param window: Window instance
        """
        self.window = window

    def apply(self, option: Dict[str, Any]):
        """
        Apply placeholders to option

        :param option: Option dict
        """
        if option['type'] == 'dict' and 'keys' in option:
            for key in option['keys']:
                item = option['keys'][key]
                if type(item) is dict:
                    if "type" in item:
                        if item["type"] == "combo":
                            if "use" in item and item["use"] is not None:
                                item["keys"] = self.apply_by_id(item["use"])
        elif option['type'] == 'cmd' and 'params_keys' in option:
            for key in option['params_keys']:
                item = option['params_keys'][key]
                if type(item) is dict:
                    if "type" in item:
                        if item["type"] == "combo":
                            if "use" in item and item["use"] is not None:
                                item["keys"] = self.apply_by_id(item["use"])
        elif option['type'] == 'combo':
            if "use" in option and option["use"] is not None:
                option["keys"] = self.apply_by_id(option["use"])

    def apply_by_id(self, id: str) -> List[Dict[str, str]]:
        """
        Apply placeholders by id

        :param id: Placeholder options id
        """
        if id == "presets":
            return self.get_presets()
        elif id == "models":
            return self.get_models()
        elif id == "langchain_providers":
            return self.get_langchain_providers()
        elif id == "llama_index_providers":
            return self.get_llama_index_providers()
        elif id == "embeddings_providers":
            return self.get_embeddings_providers()
        elif id == "llama_index_loaders":
            return self.get_llama_index_loaders()
        elif id == "llama_index_loaders_file":
            return self.get_llama_index_loaders(type="file")
        elif id == "llama_index_loaders_web":
            return self.get_llama_index_loaders(type="web")
        elif id == "llama_index_chat_modes":
            return self.get_llama_index_chat_modes()
        elif id == "vector_storage":
            return self.get_vector_storage()
        elif id == "var_types":
            return self.get_var_types()
        elif id == "agent_modes":
            return self.get_agent_modes()
        elif id == "agent_provider":
            return self.get_agent_providers()
        elif id == "syntax_styles":
            return self.get_syntax_styles()
        elif id == "styles":
            return self.get_styles()
        elif id == "idx":
            return self.get_idx()
        elif id == "keys":
            return self.get_keys()
        elif id == "keys_modifiers":
            return self.get_modifiers()
        elif id == "access_actions":
            return self.get_access_actions()
        elif id == "speech_synthesis_actions":
            return self.get_speech_synthesis_actions()
        elif id == "voice_control_actions":
            return self.get_voice_control_actions()
        elif id == "audio_input_devices":
            return self.get_audio_input_devices()
        elif id == "audio_tts_whisper_voices":
            return self.get_audio_tts_whisper_voices()
        else:
            return []

    def get_audio_tts_whisper_voices(self) -> List[Dict[str, str]]:
        """
        Get audio TTS whisper voices list

        :return: placeholders list
        """
        voices = self.window.core.audio.whisper.get_voices()
        data = []
        for voice in voices:
            data.append({voice: voice})
        return data

    def get_audio_input_devices(self) -> List[Dict[str, str]]:
        """
        Get audio input devices list

        :return: placeholders list
        """
        devices = self.window.core.audio.get_input_devices()
        data = []
        for device in devices:
            data.append({str(device[0]): device[1]})
        return data

    def get_langchain_providers(self) -> List[Dict[str, str]]:
        """
        Get Langchain LLM provider placeholders list

        :return: placeholders list
        """
        ids = self.window.core.llm.get_ids(MODE_LANGCHAIN)
        data = []
        for id in ids:
            data.append({id: id})
        return data

    def get_llama_index_providers(self) -> List[Dict[str, str]]:
        """
        Get Llama-index LLM provider placeholders list

        :return: placeholders list
        """
        ids = self.window.core.llm.get_ids(MODE_LLAMA_INDEX)
        data = []
        for id in ids:
            data.append({id: id})
        return data

    def get_agent_providers(self) -> List[Dict[str, str]]:
        """
        Get Llama-index agent provider placeholders list

        :return: placeholders list
        """
        ids = self.window.core.agents.provider.get_providers()
        data = []
        for id in ids:
            data.append({id: id})
        return data

    def get_embeddings_providers(self) -> List[Dict[str, str]]:
        """
        Get embeddings placeholders list

        :return: placeholders list
        """
        ids = self.window.core.llm.get_ids("embeddings")
        data = []
        for id in ids:
            data.append({id: id})
        return data

    def get_llama_index_chat_modes(self) -> List[Dict[str, str]]:
        """
        Get llama chat modes list

        :return: placeholders list
        """
        return [
            {"best": "best"},
            {"condense_question": "condense_question"},
            {"context": "context (default)"},
            {"condense_plus_context": "condense_plus_context"},
            {"simple": "simple (no query engine)"},
            {"react": "react"},
            {"openai": "openai"},
        ]

    def get_llama_index_loaders(self, type: str = "all") -> List[Dict[str, str]]:
        """
        Get data loaders placeholders list

        :param type: data type
        :return: placeholders list
        """
        data = []
        choices = self.window.controller.idx.common.get_loaders_choices() # list
        for choice in choices:
            if type == "all":
                data.append(choice)
            elif type == "file":
                key = list(choice.keys())[0]
                if key.startswith("file_"):
                    data.append(choice)
            elif type == "web":
                key = list(choice.keys())[0]
                if key.startswith("web_"):
                    data.append(choice)
        return data

    def get_vector_storage(self) -> List[Dict[str, str]]:
        """
        Get vector storage placeholders list

        :return: placeholders list
        """
        ids = self.window.core.idx.storage.get_ids()
        data = []
        for id in ids:
            data.append({id: id})
        return data

    def get_var_types(self) -> List[Dict[str, str]]:
        """
        Get langchain placeholders list

        :return: placeholders list
        """
        types = ["str", "int", "float", "bool", "dict", "list", "None"]
        data = []
        for type in types:
            data.append({type: type})
        return data

    def get_presets(self) -> List[Dict[str, str]]:
        """
        Get presets placeholders list

        :return: Presets placeholders list
        """
        presets = self.window.core.presets.get_all()
        data = []
        data.append({'_': '---'})
        for id in presets:
            if id.startswith("current."):  # ignore "current" preset
                continue
            data.append({id: id})  # TODO: name
        return data

    def get_models(self) -> List[Dict[str, str]]:
        """
        Get models placeholders list

        :return: Models placeholders list
        """
        models = self.window.core.models.get_all()
        data = []
        for id in models:
            model = models[id]
            suffix = ""
            if "provider" in model.llama_index and model.llama_index["provider"] == "ollama":
                suffix = " [Ollama]"
            name = model.name + suffix
            data.append({id: name})
        return data

    def get_agent_modes(self) -> List[Dict[str, str]]:
        """
        Get agent/expert modes placeholders list

        :return: Models placeholders list
        """
        modes = self.window.core.agents.legacy.get_allowed_modes()
        data = []
        for id in modes:
            name = trans(f"mode.{id}")
            data.append({id: name})
        return data

    def get_idx(self) -> List[Dict[str, str]]:
        """
        Get indexes placeholders list

        :return: Indexes placeholders list
        """
        indexes = self.window.core.idx.get_idx_ids()
        data = []
        data.append({'_': '---'})
        for item in indexes:
            for k, v in item.items():
                data.append({k: v})
        return data

    def get_syntax_styles(self) -> List[Dict[str, str]]:
        """
        Get highlighter styles list

        :return: placeholders list
        """
        styles = self.window.controller.chat.render.web_renderer.body.highlight.get_styles()
        styles.sort()
        data = []
        for id in styles:
            data.append({id: id})
        return data

    def get_styles(self) -> List[Dict[str, str]]:
        """
        Get styles list

        :return: placeholders list
        """
        styles = self.window.controller.theme.common.get_styles_list()
        styles.sort()
        data = []
        for id in styles:
            name = id
            data.append({id: name})
        return data

    def get_keys(self) -> List[Dict[str, str]]:
        """
        Get keys

        :return: keys
        """
        return self.window.core.access.shortcuts.get_keys_choices()

    def get_modifiers(self) -> List[Dict[str, str]]:
        """
        Get modifiers

        :return: keys
        """
        return self.window.core.access.shortcuts.get_modifiers_choices()

    def get_access_actions(self) -> List[Dict[str, str]]:
        """
        Get access actions list

        :return: app actions list
        """
        choices = self.window.core.access.actions.get_access_choices()
        translated_choices = []
        for choice in choices:
            for key, value in choice.items():
                translated_choices.append({key: trans(value)})
        # sort by translated values
        translated_choices.sort(key=lambda x: list(x.values())[0])
        return translated_choices

    def get_speech_synthesis_actions(self) -> List[Dict[str, str]]:
        """
        Get speech actions list

        :return: app actions list
        """
        choices = self.window.core.access.actions.get_speech_synthesis_choices()
        translated_choices = []
        for choice in choices:
            for key, value in choice.items():
                translated_choices.append({key: trans(value)})
        # sort by translated values
        translated_choices.sort(key=lambda x: list(x.values())[0])
        return translated_choices

    def get_voice_control_actions(self) -> List[Dict[str, str]]:
        """
        Get voice control actions list

        :return: app actions list
        """
        choices = self.window.core.access.actions.get_voice_control_choices()
        translated_choices = []
        for choice in choices:
            for key, value in choice.items():
                translated_choices.append({key: trans(value)})
        # sort by translated values
        translated_choices.sort(key=lambda x: list(x.values())[0])
        return translated_choices





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.18 03:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QPainter, QPalette
from PySide6.QtWidgets import QWidget

import pygpt_net.icons_rc

class InputBar(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._level = 0.0  # level from 0.0 to 100.0
        self.setFixedSize(200, 5)  # bar size

    def setLevel(self, level):
        """
        Set volume level

        :param level: level
        """
        self._level = level
        self.update()

    def paintEvent(self, event):
        """
        Paint event

        :param event: event
        """
        palette = self.palette()
        painter = QPainter(self)
        painter.fillRect(self.rect(), Qt.transparent)
        level_width = (self._level / 100.0) * self.width()
        painter.setBrush(palette.color(QPalette.ButtonText))
        painter.setPen(Qt.NoPen)
        painter.drawRect(0, 0, level_width, self.height())

    """
        # --- bar from center ---
        def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(self.rect(), Qt.transparent)
        level_width = (self._level / 100.0) * self.width()
        half_level_width = level_width / 2
        center_x = self.width() / 2
        rect_x = center_x - half_level_width
        painter.setBrush(Qt.green)
        painter.setPen(Qt.NoPen)
        painter.drawRect(rect_x, 0, level_width, self.height())
    """


class OutputBar(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._level = 0.0  # level from 0.0 to 100.0
        self.setFixedSize(200, 5)  # bar size

    def setLevel(self, level):
        """
        Set volume level

        :param level: level
        """
        self._level = level
        self.update()

    def paintEvent(self, event):
        """
        Paint event

        :param event: event
        """
        palette = self.palette()
        painter = QPainter(self)
        painter.fillRect(self.rect(), Qt.transparent)
        level_width = (self._level / 100.0) * self.width()
        painter.setBrush(palette.color(QPalette.ButtonText))
        painter.setPen(Qt.NoPen)
        painter.drawRect(0, 0, level_width, self.height())

    """
        # --- bar from center ---
        def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(self.rect(), Qt.transparent)
        level_width = (self._level / 100.0) * self.width()
        half_level_width = level_width / 2
        center_x = self.width() / 2
        rect_x = center_x - half_level_width
        painter.setBrush(Qt.green)
        painter.setPen(Qt.NoPen)
        painter.drawRect(rect_x, 0, level_width, self.height())
    """


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 21:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.plugin import BasePlugin


class BaseProvider:
    def __init__(self, plugin=None):
        """
        Audio input base provider

        :param plugin: plugin instance
        """
        self.plugin = plugin
        self.id = ""  # unique provider id
        self.name = ""  # name to display

    def init(self, plugin: BasePlugin):
        """
        Initialize provider

        :param plugin: plugin instance
        """
        self.attach(plugin)
        self.init_options()

    def attach(self, plugin: BasePlugin):
        """
        Attach plugin instance

        :param plugin: plugin instance
        """
        self.plugin = plugin

    def init_options(self):
        """Initialize provider options (for plugin settings)"""
        pass

    def transcribe(self, path: str) -> str:
        """
        Audio to text transcription

        :param path: path to audio file to transcribe
        :return: transcribed text
        """
        pass

    def is_configured(self) -> bool:
        """
        Check if provider is configured

        :return: True if configured, False otherwise
        """
        pass

    def get_config_message(self) -> str:
        """
        Return message to display when provider is not configured

        :return: message
        """
        return ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import json

from llama_index.core.readers.base import BaseReader

from .hub.github.issues import GitHubRepositoryIssuesReader
from .base import BaseLoader


class Loader(BaseLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = "github_issues"
        self.name = "GitHub Issues"
        self.type = ["web"]
        self.instructions = [
            {
                "github_issues": {
                    "description": "read and index GitHub issues",
                    "args": {
                        "owner": {
                            "type": "str",
                            "label": "Owner",
                        },
                        "repository": {
                            "type": "str",
                            "label": "Repository",
                        },
                        "state": {
                            "type": "enum",
                            "options": ["open", "closed", "all"],
                            "label": "State",
                            "description": "Issue state, default: open, available options: open, closed, all",
                        },
                        "label_filters_include": {
                            "type": "list",
                            "label": "Label filters include",
                            "description": "List of labels to include, separated by comma (,)",
                        },
                        "label_filters_exclude": {
                            "type": "list",
                            "label": "Label filters exclude",
                            "description": "List of labels to exclude, separated by comma (,)",
                        },
                    },
                }
            }
        ]
        self.init_args = {
            "token": "",
            "verbose": False,
        }
        self.init_args_types = {
            "token": "str",
            "verbose": "bool",
        }

    def get(self) -> BaseReader:
        """
        Get reader instance

        :return: Data reader instance
        """
        args = self.get_args()
        return GitHubRepositoryIssuesReader(**args)

    def get_external_id(self, args: dict = None) -> str:
        """
        Get unique web content identifier

        :param args: load_data args
        :return: unique content identifier
        """
        unique = {}
        if "owner" in args and args.get("owner"):
            unique["owner"] = args.get("owner")
        if "repository" in args and args.get("repository"):
            unique["repository"] = args.get("repository")
        if "state" in args and args.get("state"):
            unique["state"] = args.get("state")
        return json.dumps(unique)

    def prepare_args(self, **kwargs) -> dict:
        """
        Prepare arguments for load_data() method

        :param kwargs: keyword arguments
        :return: args to pass to reader
        """
        args = {}

        # repository
        if "owner" in kwargs and kwargs.get("owner"):
            if isinstance(kwargs.get("owner"), str):
                args["owner"] = kwargs.get("owner")  # repo owner
        if "repository" in kwargs and kwargs.get("repository"):
            if isinstance(kwargs.get("repository"), str):
                args["repository"] = kwargs.get("repository")  # repo name
        if "state" in kwargs and kwargs.get("state"):
            if isinstance(kwargs.get("state"), str):
                args["state"] = kwargs.get("state")
                if args["state"] not in ["open", "closed", "all"]:
                    args["state"] = "open"

        # filters
        if "label_filters_include" in kwargs and kwargs.get("label_filters_include"):
            if isinstance(kwargs.get("label_filters_include"), list):
                args["label_filters_include"] = kwargs.get("label_filters_include")
            elif isinstance(kwargs.get("label_filters_include"), str):
                args["label_filters_include"] = self.explode(kwargs.get("label_filters_include"))
        if "label_filters_exclude" in kwargs and kwargs.get("label_filters_exclude"):
            if isinstance(kwargs.get("label_filters_exclude"), list):
                args["label_filters_exclude"] = kwargs.get("label_filters_exclude")
            elif isinstance(kwargs.get("label_filters_exclude"), str):
                args["label_filters_exclude"] = self.explode(kwargs.get("label_filters_exclude"))

        return args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.02 02:00:00                  #
# ================================================== #

import requests

class Ollama:
    def __init__(self, window=None):
        """
        Ollama core

        :param window: Window instance
        """
        self.window = window
        self.available_models = []

    def get_status(self) -> dict:
        """
        Check Ollama status

        :return: dict
        """
        url = "http://localhost:11434/api/tags"
        try:
            response = requests.get(url, timeout=2)
            if response.status_code == 200:
                data = response.json()
                return {
                    'status': True,
                    'models': data.get('models', [])
                }
            else:
                return {
                    'status': False,
                    'models': []
                }
        except requests.exceptions.RequestException:
            return {
                'status': False,
                'models': []
            }

    def check_model(self, model: str) -> dict:
        """
        Check if model is available

        :param model: model ID
        :return: dict
        """
        result = {
            'is_installed': False,
            'is_model': False,
        }
        if model in self.available_models:
            return {
                'is_installed': True,
                'is_model': True,
            }
        status = self.get_status()
        if not status.get('status'):
            return result
        if "models" not in status:
            return result
        for item in status.get('models', []):
            model_id = item.get('name').replace(":latest", "")
            if model_id not in self.available_models:
                self.available_models.append(model_id)
            if model_id == model:
                return {
                    'is_installed': True,
                    'is_model': True,
                }
        return {
            'is_installed': True,
            'is_model': False,
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.29 07:00:00                  #
# ================================================== #

import re

from bs4 import BeautifulSoup


def output_html2text(html: str) -> str:
    """
    Convert output HTML to plain text

    :param html: HTML content
    :return: Plain text
    """
    if html == "":
        return ""
    try:
        soup = BeautifulSoup(html, 'html.parser')
        # remove headers from code blocks
        for tag in soup.find_all('p', class_='code-header-wrapper'):
            empty = soup.new_tag('p')
            empty.string = '\n'
            tag.replace_with(empty)
        # add separators
        for tag in soup.find_all('div', class_='msg-bot'):
            sep = soup.new_tag('p')
            sep.string = '\n\n'
            tag.insert_before(sep)
        for tag in soup.find_all('div', class_='msg-user'):
            sep = soup.new_tag('p')
            sep.string = '\n\n'
            tag.insert_before(sep)
        return soup.get_text()
    except Exception as e:
        pass
    return ""

def output_clean_html(html: str) -> str:
    """
    Clean output HTML content

    :param html: HTML content
    :return: HTML content
    """
    try:
        soup = BeautifulSoup(html, 'html.parser')
        # remove copy to clipboard from code blocks
        for tag in soup.find_all('a', class_='code-header-copy'):
            tag.decompose()
        # remove action icons
        for tag in soup.find_all('div', class_='action-icons'):
            tag.decompose()
        return str(soup)
    except Exception as e:
        pass
    return html

def has_unclosed_code_tag(text: str) -> bool:
    """
    Check if HTML content has unclosed code block

    :param text: HTML content
    :return: True if unclosed code block found
    """
    if text is None:
        return False
    if re.search(r'```(?!.*```)', text):
        return True
    return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, List, Dict

from pygpt_net.utils import trans


class Body:

    def __init__(self, window=None):
        """
        Plain-text body

        :param window: Window instance
        """
        self.window = window

    def get_image_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get image HTML

        :param url: URL to image
        :param num: number of image
        :param num_all: number of all images
        :return: HTML code
        """
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        url, path = self.window.core.filesystem.extract_local_url(url)
        return """\n{prefix}{num}: {path}\n""". \
            format(prefix=trans('chat.prefix.img'),
                   path=path,
                   num=num_str)

    def get_url_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get URL HTML

        :param url: external URL
        :param num: number of URL
        :param num_all: number of all URLs
        :return: HTML
        """
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        return """{prefix}{num}: {url}""". \
            format(prefix=trans('chat.prefix.url'),
                   url=url,
                   num=num_str)

    def get_docs_html(self, docs: List[Dict]) -> str:
        """
        Get Llama-index doc metadata HTML

        :param docs: list of document metadata
        :return: HTML code
        """
        html = ""
        html_sources = ""
        num = 1
        max = 3
        try:
            for item in docs:
                for uuid, doc_json in item.items():
                    """
                    Example doc (file metadata):
                    {'a2c7af6d-3c34-4c28-bf2d-6161e7fb525e': {
                        'file_path': '/home/user/.config/pygpt-net/data/my_cars.txt',
                        'file_name': '/home/user/.config/pygpt-net/data/my_cars.txt', 'file_type': 'text/plain',
                        'file_size': 28, 'creation_date': '2024-03-03', 'last_modified_date': '2024-03-03',
                        'last_accessed_date': '2024-03-03'}}
                    """
                    doc_parts = []
                    for key in doc_json:
                        doc_parts.append("{}: {}".format(key, doc_json[key]))
                    html_sources += "\n[{}] {}: {}".format(num, uuid, ", ".join(doc_parts))
                    num += 1
                if num >= max:
                    break
        except Exception as e:
            pass

        if html_sources != "":
            html += "\n----------\n{prefix}:\n".format(prefix=trans('chat.prefix.doc'))
            html += html_sources
        return html

    def get_file_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get file HTML

        :param url: URL to file
        :param num: number of file
        :param num_all: number of all files
        :return: HTML
        """
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        url, path = self.window.core.filesystem.extract_local_url(url)
        return """\n{prefix}{num}: {path}\n""". \
            format(prefix=trans('chat.prefix.file'),
                   path=path,
                   num=num_str)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.02 02:00:00                  #
# ================================================== #

from typing import Optional, Any, Dict

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.bridge.context import MultimodalContext
from pygpt_net.core.events import Event, AppEvent, KernelEvent, RenderEvent
from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_LLAMA_INDEX,
    MODE_ASSISTANT,
    MODE_IMAGE,
)
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Input:
    def __init__(self, window=None):
        """
        Input controller

        :param window: Window instance
        """
        self.window = window
        self.locked = False
        self.stop = False
        self.generating = False
        self.no_ctx_idx_modes = [
            # MODE_IMAGE,
            MODE_ASSISTANT,
            # MODE_LLAMA_INDEX,
            MODE_AGENT,
        ]  # assistant is handled in async, agent is handled in agent flow
        self.stop_commands = [
            "stop",
            "halt",
        ]

    def send_input(self, force: bool = False):
        """
        Send text from user input (called from UI)

        :param force: force send
        """
        self.window.controller.agent.experts.unlock()  # unlock experts
        self.window.controller.agent.llama.reset_eval_step()  # reset evaluation steps
        self.window.controller.ui.tabs.switch_to_first_chat()  # switch to first active chat tab

        # get text from input
        text = self.window.ui.nodes['input'].toPlainText().strip()
        mode = self.window.core.config.get('mode')

        if not force:
            self.window.dispatch(AppEvent(AppEvent.INPUT_SENT))  # app event

            # check if not in edit mode
            if self.window.controller.ctx.extra.is_editing():
                self.window.controller.ctx.extra.edit_submit()
                return

            # if agent mode: iterations check, show alert confirm if infinity loop
            if self.window.controller.agent.common.is_infinity_loop(mode):
                self.window.controller.agent.common.display_infinity_loop_confirm()
                return

            # check ollama model
            model = self.window.core.config.get('model')
            if mode == MODE_LLAMA_INDEX and model is not None:
                model_data = self.window.core.models.get(model)
                if model_data is not None and model_data.is_ollama():
                    model_id = model_data.get_ollama_model()
                    status = self.window.core.models.ollama.check_model(model_id)
                    is_installed = status.get('is_installed', False)
                    is_model = status.get('is_model', False)
                    if not is_installed:
                        self.window.ui.dialogs.alert(trans("dialog.ollama.not_installed"))
                        return
                    if not is_model:
                        self.window.ui.dialogs.alert(
                            trans("dialog.ollama.model_not_found").replace("{model}", model_id))
                        return

        # listen for stop command
        if self.generating \
                and text is not None \
                and text.lower().strip() in self.stop_commands:
            self.window.controller.kernel.stop()  # TODO: to chat main
            self.window.dispatch(RenderEvent(RenderEvent.CLEAR_INPUT))
            return

        # agent modes
        if mode == MODE_AGENT:
            self.window.controller.agent.legacy.on_user_send(text)  # begin legacy agent flow
        elif mode == MODE_AGENT_LLAMA:
            self.window.controller.agent.llama.on_user_send(text)  # begin llama agent flow

        # event: user input send (manually)
        event = Event(Event.USER_SEND, {
            'value': text,
        })
        self.window.dispatch(Event(Event.USER_SEND, {
            'value': text,
        }))
        text = event.data['value']

        # handle attachments with additional context (not images here)
        if mode != MODE_ASSISTANT and self.window.controller.chat.attachment.has(mode):
            self.window.dispatch(KernelEvent(KernelEvent.STATE_BUSY, {
                "id": "chat",
                "msg": "Reading attachments..."
            }))
            try:
                self.window.controller.chat.attachment.handle(mode, text)
                return  # return here, will be handled in signal
            except Exception as e:
                self.window.dispatch(KernelEvent(KernelEvent.STATE_ERROR, {
                    "id": "chat",
                    "msg": "Error reading attachments: {}".format(str(e))
                }))
                return

        # event: handle input
        context = BridgeContext()
        context.prompt = text
        event = KernelEvent(KernelEvent.INPUT_USER, {
            'context': context,
            'extra': {},
        })
        self.window.dispatch(event)

    def send(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
    ):
        """
        Send input wrapper

        :param context: bridge context
        :param extra: extra data
        """
        text = str(context.prompt)
        prev_ctx = context.ctx
        force = extra.get("force", False)
        reply = extra.get("reply", False)
        internal = extra.get("internal", False)
        parent_id = extra.get("parent_id", None)
        multimodal_ctx = context.multimodal_ctx
        self.execute(
            text=text,
            force=force,
            reply=reply,
            internal=internal,
            prev_ctx=prev_ctx,
            parent_id=parent_id,
            multimodal_ctx=multimodal_ctx,
        )

    def execute(
            self,
            text: str,
            force: bool = False,
            reply: bool = False,
            internal: bool = False,
            prev_ctx: Optional[CtxItem] = None,
            parent_id: Optional[int] = None,
            multimodal_ctx: Optional[MultimodalContext] = None,
    ):
        """
        Execute send input text to API

        :param text: input text
        :param force: force send (ignore input lock)
        :param reply: reply mode (from plugins)
        :param internal: internal call
        :param prev_ctx: previous context (if reply)
        :param parent_id: parent id (if expert)
        :param multimodal_ctx: multimodal context
        """
        self.window.dispatch(KernelEvent(KernelEvent.STATE_IDLE, {
            "id": "chat",
        }))

        # check if input is not locked
        if self.locked and not force and not internal:
            return

        self.log("Begin.")
        self.generating = True  # set generating flag

        mode = self.window.core.config.get('mode')
        if mode == MODE_ASSISTANT:
            # check if assistant is selected
            if self.window.core.config.get('assistant') is None \
                    or self.window.core.config.get('assistant') == "":
                self.window.ui.dialogs.alert(trans('error.assistant_not_selected'))
                self.generating = False  # unlock
                return
        elif self.window.controller.ui.vision.has_vision():
            # handle auto capture
            self.window.controller.camera.handle_auto_capture()

        # unlock Assistants run thread if locked
        self.window.controller.assistant.threads.stop = False
        self.window.controller.kernel.resume()

        self.log("Input prompt: {}".format(text))  # log

        # agent mode
        if mode == MODE_AGENT:
            self.log("Agent: input before: {}".format(text))
            text = self.window.controller.agent.legacy.on_input_before(text)

        # event: before input
        event = Event(Event.INPUT_BEFORE, {
            'value': text,
            'mode': mode,
        })
        self.window.dispatch(event)
        text = event.data['value']

        # check if image captured from camera, # check if attachment exists
        camera_captured = (self.window.controller.ui.vision.has_vision()
                           and self.window.controller.attachment.has(mode))

        # allow empty text input only if multimodal data, otherwise abort
        is_audio = multimodal_ctx is not None and multimodal_ctx.is_audio_input
        if len(text.strip()) == 0 and (not camera_captured and not is_audio):
            self.generating = False  # unlock as not generating
            return

        # check API key, show monit if no API key
        if mode not in self.window.controller.launcher.no_api_key_allowed:
            if not self.window.controller.chat.common.check_api_key():
                self.generating = False
                self.window.dispatch(KernelEvent(KernelEvent.STATE_ERROR, {
                    "id": "chat",
                }))
                return

        # set state to: busy
        self.window.dispatch(KernelEvent(KernelEvent.STATE_BUSY, {
            "id": "chat",
            "msg": trans('status.sending'),
        }))

        # clear input field if clear-on-send is enabled
        if self.window.core.config.get('send_clear') and not force and not internal:
            self.window.dispatch(RenderEvent(RenderEvent.CLEAR_INPUT))

        # prepare ctx, create new ctx meta if there is no ctx, or no ctx selected
        if self.window.core.ctx.count_meta() == 0 or self.window.core.ctx.get_current() is None:
            self.window.core.ctx.new()
            self.window.controller.ctx.update()
            self.log("New context created...")  # log
        else:
            # check if current ctx is allowed for this mode - if not, then auto-create new ctx
            self.window.controller.ctx.handle_allowed(mode)

        # send input to API
        if mode == MODE_IMAGE:
            self.window.controller.chat.image.send(
                text=text,
                prev_ctx=prev_ctx,
                parent_id=parent_id,
            )  # image mode
        else:
            # async
            self.window.controller.chat.text.send(
                text=text,
                reply=reply,
                internal=internal,
                prev_ctx=prev_ctx,
                parent_id=parent_id,
                multimodal_ctx=multimodal_ctx,
            )  # text mode: OpenAI, Langchain, Llama, etc.

    def log(self, data: Any):
        """
        Log data to debug

        :param data: Data to log
        """
        self.window.controller.chat.log(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.01 11:00:00                  #
# ================================================== #

from pygpt_net.item.model import ModelItem


class Gpt:
    def __init__(self, window=None):
        """
        GPT prompt templates

        :param window: Window instance
        """
        self.window = window

    def get(
            self,
            prompt: str,
            mode: str = None,
            model: ModelItem = None
    ) -> str:
        """
        Get system prompt content

        CMD/TOOL EXECUTE prompts:
         - cmd
         - cmd.extra
         - cmd.extra.assistants

        :param prompt: id of the prompt
        :param mode: mode
        :param model: model item
        :return: text content
        """
        key = "prompt." + prompt
        if self.window.core.config.has(key):
            return str(self.window.core.config.get(key))
        return ""


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import os
from typing import Optional, List, Dict

from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Body:

    def __init__(self, window=None):
        """
        Markdown Body

        :param window: Window instance
        """
        self.window = window
        self.img_width = 400

    def get_image_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get image HTML

        :param url: URL to image
        :param num: number of image
        :param num_all: number of all images
        :return: HTML code
        """
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        url, path = self.window.core.filesystem.extract_local_url(url)
        return """<a href="{url}"><img src="{path}" width="{img_width}" class="image"></a>
        <p><b>{prefix}{num}:</b> <a href="{url}">{path}</a></p>""". \
            format(prefix=trans('chat.prefix.img'),
                   url=url,
                   path=path,
                   img_width=self.img_width,
                   num=num_str)

    def get_url_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get URL HTML

        :param url: external URL
        :param num: number of URL
        :param num_all: number of all URLs
        :return: HTML code
        """
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        return """<b>{prefix}{num}:</b> <a href="{url}">{url}</a>""". \
            format(prefix=trans('chat.prefix.url'),
                   url=url,
                   num=num_str)

    def get_docs_html(self, docs: List[Dict]) -> str:
        """
        Get Llama-index doc metadata HTML

        :param docs: list of document metadata
        :return: HTML code
        """
        html = ""
        html_sources = ""
        num = 1
        for doc_json in docs:
            try:
                """
                Example doc (file metadata):
                {'a2c7af6d-3c34-4c28-bf2d-6161e7fb525e': {
                    'file_path': '/home/user/.config/pygpt-net/data/my_cars.txt',
                    'file_name': '/home/user/.config/pygpt-net/data/my_cars.txt', 'file_type': 'text/plain',
                    'file_size': 28, 'creation_date': '2024-03-03', 'last_modified_date': '2024-03-03',
                    'last_accessed_date': '2024-03-03'}}
                """
                doc_id = list(doc_json.keys())[0]
                doc_parts = []
                for key in doc_json[doc_id]:
                    doc_parts.append("<b>{}:</b> {}".format(key, doc_json[doc_id][key]))
                html_sources += "[{}] {}: {}".format(num, doc_id, ", ".join(doc_parts))
                num += 1
            except Exception as e:
                pass

        if html_sources != "":
            html += "<p><b>{prefix}:</b></p>".format(prefix=trans('chat.prefix.doc'))
            html += "<div class=\"cmd\">"
            html += "<p>" + html_sources + "</p>"
            html += "</div> "
        return html

    def get_file_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get file HTML

        :param url: URL to file
        :param num: number of file
        :param num_all: number of all files
        :return: HTML code
        """
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        url, path = self.window.core.filesystem.extract_local_url(url)
        return """<div><b>{prefix}{num}:</b> <a href="{url}">{path}</a></div>""". \
            format(prefix=trans('chat.prefix.file'),
                   url=url,
                   path=path,
                   num=num_str)

    def get_action_icons(
            self,
            ctx: CtxItem,
            all: bool = False
    ) -> list:
        """
        Get action icons for context item

        :param ctx: context item
        :param all: True to show all icons
        :return: list of icons
        """
        icons = []

        # audio read
        if ctx.output is not None and ctx.output != "":
            icons.append(
                '<a href="extra-audio-read:{}"><span class="cmd">{}</span></a>'.format(
                    ctx.id,
                    self.get_icon("audio", trans("ctx.extra.audio"))
                )
            )
            # copy ctx
            icons.append(
                '<a href="extra-copy:{}"><span class="cmd">{}</span></a>'.format(
                    ctx.id,
                    self.get_icon("copy", trans("ctx.extra.copy"))
                )
            )
            # regen link
            icons.append(
                '<a href="extra-replay:{}"><span class="cmd">{}</span></a>'.format(
                    ctx.id,
                    self.get_icon("reload", trans("ctx.extra.reply"))
                )
            )
            if all:
                # edit link
                icons.append(
                    '<a href="extra-edit:{}"><span class="cmd">{}</span></a>'.format(
                        ctx.id,
                        self.get_icon("edit", trans("ctx.extra.edit"))
                    )
                )
                # delete link
                icons.append(
                    '<a href="extra-delete:{}"><span class="cmd">{}</span></a>'.format(
                        ctx.id,
                        self.get_icon("delete", trans("ctx.extra.delete"))
                    )
                )

                # join link
                if not self.window.core.ctx.is_first_item(ctx.id):
                    icons.append(
                        '<a href="extra-join:{}"><span class="cmd">{}</span></a>'.format(
                            ctx.id,
                            self.get_icon("join", trans("ctx.extra.join"))
                        )
                    )
        return icons

    def get_icon(self, icon: str, title: Optional[str] = None) -> str:
        """
        Get icon

        :param icon: icon name
        :param title: icon title
        :return: icon HTML
        """
        icon = os.path.join(self.window.core.config.get_app_path(), "data", "icons", "chat", icon + ".png")
        return '<img src="{}" width="20" height="20" title="{}" alt="{}">'.format(icon, title, title)
        # return '<img src=":/icons/{}.svg" width="25" title="{}">'.format(icon, title)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QSplitter, QWidget, QSizePolicy

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_IMAGE,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
)
from pygpt_net.ui.base.config_dialog import BaseConfigDialog
from pygpt_net.ui.widget.dialog.editor import EditorDialog
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.lists.experts import ExpertsEditor
from pygpt_net.utils import trans


class Preset(BaseConfigDialog):
    def __init__(self, window=None, *args, **kwargs):
        super(Preset, self).__init__(window, *args, **kwargs)
        """
        Preset editor dialog

        :param window: Window instance
        """
        self.window = window
        self.id = "preset"
        self.dialog_id = "preset.presets"

    def setup(self):
        """Setup preset editor dialog"""
        self.window.ui.nodes['preset.btn.current'] = QPushButton(trans("dialog.preset.btn.current"))
        self.window.ui.nodes['preset.btn.save'] = QPushButton(trans("dialog.preset.btn.save"))
        self.window.ui.nodes['preset.btn.current'].clicked.connect(
            lambda: self.window.controller.presets.editor.from_current()
        )
        self.window.ui.nodes['preset.btn.save'].clicked.connect(
            lambda: self.window.controller.presets.editor.save()
        )

        self.window.ui.nodes['preset.btn.current'].setAutoDefault(False)
        self.window.ui.nodes['preset.btn.save'].setAutoDefault(True)

        footer = QHBoxLayout()
        footer.addWidget(self.window.ui.nodes['preset.btn.current'])
        footer.addWidget(self.window.ui.nodes['preset.btn.save'])

        # fields
        self.window.ui.paths[self.id] = QLabel(str(self.window.core.config.path))

        # get option fields config
        fields = self.window.controller.presets.editor.get_options()

        # build settings widgets
        widgets = self.build_widgets(self.id, fields)  # from base config dialog

        # apply settings widgets
        for key in widgets:
            self.window.ui.config[self.id][key] = widgets[key]

        # btn: add function
        self.window.ui.config[self.id]['tool.function'].add_btn.setText(trans('assistant.func.add'))

        # apply widgets to layouts
        options = {}
        for key in widgets:
            if fields[key]["type"] in ['text', 'int', 'float']:
                options[key] = self.add_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'textarea':
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'bool':
                widgets[key].setMaximumHeight(38)
                options[key] = self.add_raw_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'dict':
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'combo':
                options[key] = self.add_option(widgets[key], fields[key])

        self.window.ui.nodes['preset.tool.function.label'].setVisible(False)  # hide label

        rows = QVBoxLayout()

        for key in options:
            options[key].setContentsMargins(0, 0, 0, 0)

        # modes
        mode_keys = [
            MODE_CHAT,
            MODE_AUDIO,
            MODE_COMPLETION,
            MODE_IMAGE,
            MODE_VISION,
            MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_AGENT_LLAMA,
            MODE_AGENT,
            MODE_EXPERT,
            MODE_RESEARCH,
        ]
        rows_mode = QVBoxLayout()
        rows_mode.addStretch()
        rows_mode.setContentsMargins(0, 0, 0, 0)
        for key in mode_keys:
            rows_mode.addLayout(options[key])

        # modes
        self.window.ui.nodes['preset.editor.modes'] = QWidget()
        self.window.ui.nodes['preset.editor.modes'].setLayout(rows_mode)
        self.window.ui.nodes['preset.editor.modes'].setContentsMargins(0, 0, 0, 0)
        self.window.ui.nodes['preset.editor.modes'].setMaximumWidth(300)

        # functions label
        self.window.ui.nodes['preset.tool.function.label.all'] = HelpLabel(
            trans("preset.tool.function.tip.all"))
        self.window.ui.nodes['preset.tool.function.label.all'].setAlignment(Qt.AlignCenter)
        self.window.ui.nodes['preset.tool.function.label.assistant'] = HelpLabel(
            trans("preset.tool.function.tip.assistant"))
        self.window.ui.nodes['preset.tool.function.label.assistant'].setAlignment(Qt.AlignCenter)
        self.window.ui.nodes['preset.tool.function.label.agent_llama'] = HelpLabel(
            trans("preset.tool.function.tip.agent_llama"))
        self.window.ui.nodes['preset.tool.function.label.agent_llama'].setAlignment(Qt.AlignCenter)

        # functions
        self.window.ui.nodes['preset.editor.functions'] = QWidget()
        self.window.ui.nodes['preset.editor.functions'].setLayout(options["tool.function"])

        # experts
        self.window.ui.nodes['preset.editor.experts'] = ExpertsEditor(self.window)

        # agents - llama index
        agent_keys = [
            "agent_provider",
            "idx",
            "assistant_id",
        ]
        agent_layout = QVBoxLayout()
        agent_layout.setContentsMargins(0, 0, 0, 0)
        for key in agent_keys:
            widget = QWidget()
            widget.setLayout(options[key])
            agent_layout.addWidget(widget)
        agent_layout.addStretch()
        self.window.ui.nodes['preset.editor.agent_llama'] = QWidget()
        self.window.ui.nodes['preset.editor.agent_llama'].setLayout(agent_layout)
        self.window.ui.nodes['preset.editor.agent_llama'].setContentsMargins(20, 0, 0, 30)

        # prompt
        widget_prompt = QWidget()
        widget_prompt.setLayout(options["prompt"])
        widget_prompt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # left column
        left_keys = [
            "filename",
            "name",
            "ai_name",
            "user_name",
            "model",
            "temperature",
        ]
        for key in left_keys:
            self.window.ui.nodes['preset.editor.' + key] = QWidget()
            self.window.ui.nodes['preset.editor.' + key].setLayout(options[key])
            self.window.ui.nodes['preset.editor.' + key].setContentsMargins(0, 0, 0, 0)
            rows.addWidget(self.window.ui.nodes['preset.editor.' + key])

        rows.setContentsMargins(0, 0, 0, 0)
        rows.addStretch()

        widget_base = QWidget()
        widget_base.setLayout(rows)

        func_tip_layout = QVBoxLayout()
        func_tip_layout.addWidget(self.window.ui.nodes['preset.tool.function.label.all'])
        func_tip_layout.addWidget(self.window.ui.nodes['preset.tool.function.label.assistant'])
        func_tip_layout.addWidget(self.window.ui.nodes['preset.tool.function.label.agent_llama'])
        func_tip_layout.setContentsMargins(0, 0, 0, 0)
        func_tip_widget = QWidget()
        func_tip_widget.setLayout(func_tip_layout)

        func_rows = QVBoxLayout()
        func_rows.addWidget(self.window.ui.nodes['preset.editor.functions'])
        func_rows.addWidget(self.window.ui.nodes['preset.editor.experts'])
        func_rows.addWidget(self.window.ui.nodes['preset.editor.agent_llama'])
        #func_rows.addStretch()
        func_rows.addWidget(func_tip_widget)

        func_rows.setContentsMargins(0, 0, 0, 0)
        func_widget = QWidget()
        func_widget.setLayout(func_rows)
        func_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.window.ui.nodes['preset.editor.functions'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.window.ui.nodes['preset.editor.experts'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        main = QHBoxLayout()
        main.addWidget(widget_base)
        main.addWidget(func_widget)
        main.addWidget(self.window.ui.nodes['preset.editor.modes'])
        main.setContentsMargins(0, 0, 0, 0)

        widget_main = QWidget()
        widget_main.setLayout(main)

        self.window.ui.splitters['editor.presets'] = QSplitter(Qt.Vertical)
        self.window.ui.splitters['editor.presets'].addWidget(widget_main)
        self.window.ui.splitters['editor.presets'].addWidget(widget_prompt)
        self.window.ui.splitters['editor.presets'].setStretchFactor(0, 1)
        self.window.ui.splitters['editor.presets'].setStretchFactor(1, 2)

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.splitters['editor.presets'])
        layout.addLayout(footer)

        self.window.ui.dialog['editor.' + self.dialog_id] = EditorDialog(self.window, self.dialog_id)
        self.window.ui.dialog['editor.' + self.dialog_id].setLayout(layout)
        self.window.ui.dialog['editor.' + self.dialog_id].setWindowTitle(trans('dialog.preset'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.24 22:00:00                  #
# ================================================== #

from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QCheckBox, QHBoxLayout, QWidget, QLabel

from pygpt_net.ui.widget.anims.toggles import AnimToggle
from pygpt_net.utils import trans

import pygpt_net.icons_rc

class OptionCheckbox(QWidget):
    def __init__(
            self,
            window=None,
            parent_id: str = None,
            id: str = None,
            option: dict = None,
            icon = None
    ):
        """
        Settings checkbox

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        :param icon: icon
        """
        # TODO: https://pypi.org/project/QtAwesome/

        super(OptionCheckbox, self).__init__(window)
        self.window = window
        self.id = id
        self.parent_id = parent_id
        self.option = option
        self.value = False
        self.title = ""
        self.real_time = False

        # init from option data
        if self.option is not None:
            if "label" in self.option and self.option["label"] is not None \
                    and self.option["label"] != "":
                self.title = trans(self.option["label"])
            if "value" in self.option:
                self.value = self.option["value"]
            if "real_time" in self.option:
                self.real_time = self.option["real_time"]

        # self.box = QCheckBox(self.title, self.window)
        self.box = AnimToggle()
        if self.value is not None:
            self.box.setChecked(self.value)
        self.box.stateChanged.connect(
            lambda: self.window.controller.config.checkbox.on_update(
                self.parent_id,
                self.id,
                self.option,
                self.box.isChecked()
            )
        )
        self.label = QLabel(self.title)
        self.layout = QHBoxLayout()
        self.layout.addWidget(self.box)

        # add icon if defined
        if icon is not None:
            ico = QLabel()
            pixmap = QIcon(icon).pixmap(24, 24)
            ico.setPixmap(pixmap)
            self.layout.addWidget(ico)


        self.layout.addWidget(self.label)
        self.layout.addStretch()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        # self.layout = QHBoxLayout()
        #self.layout.addWidget(self.box)

        #self.setLayout(self.layout)

    def setIcon(self, icon: str):
        """
        Set icon

        :param icon: icon
        """
        self.box.setIcon(icon)

    def setText(self, text: str):
        """
        Set label text

        :param text: text
        """
        self.label.setText(text)

    def setChecked(self, state: bool):
        """
        Set checkbox state

        :param state: state
        """
        self.box.setChecked(state)

    def isChecked(self) -> bool:
        """
        Get checkbox state

        :return: state
        """
        return self.box.isChecked()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import json
import os
from pathlib import Path
from typing import Optional, Dict
from uuid import uuid4


class Profile:

    PROFILE_FILE = 'profile.json'

    def __init__(self, window=None):
        """
        Profile core

        :param window: Window instance
        """
        self.window = window
        self.base_workdir = None
        self.initialized = False
        self.version = '1.0.0'
        self.current = None
        self.profiles = {}

    def init(self, workdir: str):
        """
        Initialize profile

        :param workdir: base workdir (in user's home, real path)
        """
        self.set_base_workdir(workdir)
        self.install()
        if not self.initialized:
            self.load()
            self.update_path()

    def set_base_workdir(self, workdir: str):
        """
        Set base workdir

        :param workdir: base workdir
        """
        self.base_workdir = workdir

    def has_file(self) -> bool:
        """
        Check if profile config exists

        :return: True if exists
        """
        return os.path.exists(os.path.join(self.base_workdir, self.PROFILE_FILE))

    def from_defaults(self):
        """Create default profile config"""
        is_test = os.environ.get('ENV_TEST') == '1'
        if is_test:
            return # DISABLE in tests!!!

        uuid = str(uuid4())
        path = self.base_workdir.replace("%HOME%", str(Path.home()))

        # from current path cfg
        path_file = "path.cfg"
        p = os.path.join(self.base_workdir, path_file)
        if os.path.exists(p):
            with open(p, 'r', encoding='utf-8') as f:
                prev_path = f.read().strip()
                if prev_path != "":
                    prev_path = prev_path.replace("%HOME%", str(Path.home()))
                    if os.path.exists(prev_path):
                        path = prev_path # use previously used path

        # create default profile
        self.current = uuid
        self.profiles = {}
        self.profiles[uuid] = {
            'name': 'Default',
            'workdir': path.replace(str(Path.home()), "%HOME%")
        }
        self.save()

    def install(self):
        """Install profiles"""
        if not self.has_file():
            self.from_defaults()

    def load(self):
        """Load profiles"""
        f = os.path.join(self.base_workdir, self.PROFILE_FILE)
        if os.path.exists(f):
            with open(f, 'r', encoding='utf-8') as f:
                try:
                    data = json.load(f)
                    self.current = data['current']
                    self.profiles = data['profiles']
                    self.initialized = True
                except Exception as e:
                    print('CRITICAL: Error loading profile file:', e)

    def save(self):
        """Save profiles"""
        config = {
            '_version': self.version,
            'current': self.current,
            'profiles': self.profiles
        }
        json_data = json.dumps(config, indent=4)
        f = os.path.join(self.base_workdir, self.PROFILE_FILE)
        with open(f, 'w', encoding='utf-8') as f:
            f.write(json_data)

    def add(
            self,
            name: str,
            workdir: str
    ) -> str:
        """
        Add new profile

        :param name: profile name
        :param workdir: profile workdir
        :return: new profile uuid
        """
        uuid = str(uuid4())
        self.profiles[uuid] = {
            'name': name,
            'workdir': workdir.replace(str(Path.home()), "%HOME%")
        }
        self.save()
        return uuid

    def remove(self, uuid: str) -> bool:
        """
        Remove profile

        :param uuid: profile uuid
        :return: True if removed
        """
        if uuid in self.profiles:
            del self.profiles[uuid]
            self.save()
            return True
        return False

    def set_current(self, uuid: str) -> bool:
        """
        Set current profile

        :param uuid: profile uuid
        :return: True if set
        """
        if uuid in self.profiles:
            self.current = uuid
            self.save()
            self.update_path()
            return True
        return False

    def update_profile(
            self,
            uuid: str,
            name: str,
            workdir: str
    ) -> bool:
        """
        Update profile

        :param uuid: profile uuid
        :param name: profile name
        :param workdir: profile workdir
        :return: True if updated
        """
        if uuid in self.profiles:
            self.profiles[uuid] = {
                'name': name,
                'workdir': workdir.replace(str(Path.home()), "%HOME%")
            }
            self.save()
            return True
        return False

    def get_current(self) -> str:
        """
        Get current profile

        :return: current profile uuid
        """
        return self.current

    def get(self, uuid: str) -> Optional[dict]:
        """
        Get profile by uuid

        :param uuid: profile uuid
        :return: profile dict
        """
        if uuid in self.profiles:
            return self.profiles[uuid]
        return None

    def get_all(self) -> Dict[str, Dict]:
        """
        Get all profiles

        :return: all profiles dict
        """
        if not self.profiles:
            return {}
        sorted_profiles = sorted(self.profiles.items(), key=lambda item: item[1]['name'])
        return dict(sorted_profiles)

    def get_current_workdir(self) -> str:
        """
        Get current workdir

        :return: current workdir (real path)
        """
        if self.current is not None and self.current in self.profiles:
            tmp_dir = self.profiles[self.current]['workdir'].replace("%HOME%", str(Path.home()))
            if os.path.exists(tmp_dir):
                return tmp_dir
        return self.base_workdir

    def append(
            self,
            uuid: str,
            profile: Dict
    ):
        """
        Append profile

        :param uuid: profile uuid
        :param profile: profile dict
        """
        self.profiles[uuid] = profile
        self.save()

    def get_current_name(self) -> str:
        """
        Get current profile name

        :return: current profile name
        """
        if self.current is not None and self.current in self.profiles:
            return self.profiles[self.current]['name']
        return 'no profile'

    def update_current_workdir(self, path: str):
        """
        Update current workdir

        :param path: new workdir
        """
        if self.current is not None and self.current in self.profiles:
            self.profiles[self.current]['workdir'] = path.replace(str(Path.home()), "%HOME%")
            self.save()

    def update_path(self):
        """Update path.cfg of currently active profile"""
        path = self.get_current_workdir()
        if not os.path.exists(path):
            return  # abort if path does not exist
        if path == self.base_workdir:
            path = "" # default path
        path = path.replace(str(Path.home()), "%HOME%")
        path_file = "path.cfg"
        p = os.path.join(self.base_workdir, path_file)
        with open(p, 'w', encoding='utf-8') as f:
            f.write(path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import json

from llama_index.core.readers.base import BaseReader

from llama_index.readers.google.drive.base import GoogleDriveReader
from .base import BaseLoader


class Loader(BaseLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = "google_drive"
        self.name = "Google Drive"
        self.type = ["web"]
        self.instructions = [
            {
                "google_drive": {
                    "description": "read and index files from Google Drive",
                    "args": {
                        "folder_id": {
                            "type": "str",
                            "label": "Folder ID",
                        },
                        "file_ids": {
                            "type": "list",
                            "label": "File IDs",
                            "description": "List of file ids, separated by comma (,)",
                        },
                        "mime_types": {
                            "type": "list",
                            "label": "Mime Types",
                            "description": "List of mime types, separated by comma (,)",
                        },
                    },
                }
            }
        ]
        self.init_args = {
            "client_config": {},
            "credentials_path" : "credentials.json",
            "token_path": "token.json",
            "pydrive_creds_path": "creds.txt",
        }
        self.init_args_types = {
            "client_config": "dict",
            "credentials_path": "str",
            "token_path": "str",
            "pydrive_creds_path": "str",
        }

    def get(self) -> BaseReader:
        """
        Get reader instance

        :return: Data reader instance
        """
        args = self.get_args()
        return GoogleDriveReader(**args)

    def get_external_id(self, args: dict = None) -> str:
        """
        Get unique web content identifier

        :param args: load_data args
        :return: unique content identifier
        """
        unique = {}
        if "folder_id" in args and args.get("folder_id"):
            unique["folder_id"] = args.get("folder_id")
        if "file_ids" in args and args.get("file_ids"):
            unique["file_ids"] = args.get("file_ids")
        if "mime_types" in args and args.get("mime_types"):
            unique["mime_types"] = args.get("mime_types")
        return json.dumps(unique)

    def prepare_args(self, **kwargs) -> dict:
        """
        Prepare arguments for load_data() method

        :param kwargs: keyword arguments
        :return: args to pass to reader
        """
        args = {}
        if "url" in kwargs and kwargs.get("url"):
            if isinstance(kwargs.get("url"), str):
                args["folder_id"] = kwargs.get("url")  # default folder id

        if "folder_id" in kwargs and kwargs.get("folder_id"):
            if isinstance(kwargs.get("folder_id"), str):
                args["folder_id"] = kwargs.get("folder_id")  # folder id

        if "file_ids" in kwargs and kwargs.get("file_ids"):
            if isinstance(kwargs.get("file_ids"), list):
                args["file_ids"] = kwargs.get("file_ids")  # list of file ids
            elif isinstance(kwargs.get("file_ids"), str):
                args["file_ids"] = self.explode(kwargs.get("file_ids"))

        if "mime_types" in kwargs and kwargs.get("mime_types"):
            if isinstance(kwargs.get("mime_types"), list):
                args["mime_types"] = kwargs.get("mime_types")  # list of mime types
            elif isinstance(kwargs.get("mime_types"), str):
                args["mime_types"] = self.explode(kwargs.get("mime_types"))
        return args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, Slot, QUrl, QObject, Signal
from PySide6.QtWidgets import QVBoxLayout, QCheckBox, QHBoxLayout

from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.textarea.editor import BaseCodeEditor
from pygpt_net.ui.widget.textarea.html import HtmlOutput, CustomWebEnginePage

import pygpt_net.icons_rc
from pygpt_net.utils import trans


class ToolWidget:
    def __init__(self, window=None, tool=None):
        """
        HTML/JS canvas widget

        :param window: Window instance
        :param tool: Tool instance
        """
        self.window = window
        self.tool = tool  # tool instance
        self.output = None  # canvas output
        self.edit = None  # canvas edit
        self.btn_edit = None  # edit checkbox

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.output.set_tab(tab)
        self.edit.set_tab(tab)

    def setup(self) -> QVBoxLayout:
        """
        Setup widget body

        :return: QVBoxLayout
        """
        self.output = CanvasOutput(self.window)
        self.output.setPage(
            CustomWebEnginePage(self.window, self.output)
        )
        self.edit = CanvasEdit(self.window)
        self.edit.setVisible(False)
        self.edit.textChanged.connect(
            lambda: self.tool.save_output()
        )

        # edit checkbox
        self.btn_edit = QCheckBox(trans("html_canvas.btn.edit"))
        self.btn_edit.stateChanged.connect(
            lambda: self.tool.toggle_edit(self)
        )

        path = self.tool.get_current_path()
        path_label = HelpLabel(path)
        path_label.setMaximumHeight(30)
        path_label.setAlignment(Qt.AlignRight)

        bottom_layout = QHBoxLayout()
        bottom_layout.addWidget(self.btn_edit)
        bottom_layout.addWidget(path_label)

        output_layout = QVBoxLayout()
        output_layout.addWidget(self.output)
        output_layout.addWidget(self.edit)
        output_layout.setContentsMargins(0, 0, 0, 0)

        # connect signals
        self.output.signals.save_as.connect(
            self.tool.handle_save_as)
        self.output.signals.audio_read.connect(
            self.window.controller.chat.render.handle_audio_read)

        self.tool.signals.update.connect(self.set_output)
        self.tool.signals.reload.connect(self.load_output)

        layout = QVBoxLayout()
        layout.addLayout(output_layout)
        layout.addLayout(bottom_layout)
        return layout

    @Slot(str)
    def set_output(self, content: str):
        """
        Set output content

        :param content: Content
        """
        self.edit.setPlainText(content)

    @Slot(str)
    def load_output(self, path: str):
        """
        Load output content

        :param path: Content
        """
        self.output.setUrl(QUrl().fromLocalFile(path))

class CanvasOutput(HtmlOutput):
    def __init__(self, window=None):
        """
        HTML canvas output

        :param window: main window
        """
        super(CanvasOutput, self).__init__(window)
        self.window = window

class CanvasEdit(BaseCodeEditor):
    def __init__(self, window=None):
        """
        Python interpreter output

        :param window: main window
        """
        super(CanvasEdit, self).__init__(window)
        self.window = window
        self.setReadOnly(False)
        self.value = 12
        self.max_font_size = 42
        self.min_font_size = 8
        self.setProperty('class', 'interpreter-output')
        self.default_stylesheet = ""
        self.setStyleSheet(self.default_stylesheet)
        self.tab = None
        self.installEventFilter(self)

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == event.Type.FocusIn:
            if self.tab is not None:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)


class ToolSignals(QObject):
    update = Signal(str)  # data
    reload = Signal(str)  # path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenuBar

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.tools.code_interpreter.ui.widgets import ToolWidget
from pygpt_net.ui.widget.dialog.base import BaseDialog
from pygpt_net.utils import trans

class Tool:
    def __init__(self, window=None, tool=None):
        """
        Python interpreter dialog

        :param window: Window instance
        :param tool: Tool instance
        """
        self.window = window
        self.tool = tool  # tool instance
        self.layout = None
        self.widget = ToolWidget(window, tool)
        self.menu_bar = None
        self.menu = {}
        self.actions = {}  # menu actions

    def set_tab(self, tab: Tab):
        """
        Set tab

        :param tab: Tab
        """
        self.widget.set_tab(tab)

    def setup_menu(self) -> QMenuBar:
        """
        Setup dialog menu

        :return: QMenuBar
        """
        # create menu bar
        self.menu_bar = QMenuBar()
        self.menu["file"] = self.menu_bar.addMenu(trans("interpreter.menu.file"))
        self.menu["kernel"] = self.menu_bar.addMenu(trans("interpreter.menu.kernel"))

        self.actions["file.clear_output"] = QAction(QIcon(":/icons/close.svg"),
                                                    trans("interpreter.menu.file.clear_output"))
        self.actions["file.clear_output"].triggered.connect(
            lambda: self.tool.clear_output()
        )
        self.actions["file.clear_history"] = QAction(QIcon(":/icons/close.svg"),
                                                    trans("interpreter.menu.file.clear_history"))
        self.actions["file.clear_history"].triggered.connect(
            lambda: self.tool.clear_history()
        )
        self.actions["file.clear_all"] = QAction(QIcon(":/icons/close.svg"),
                                                   trans("interpreter.menu.file.clear_all"))
        self.actions["file.clear_all"].triggered.connect(
            lambda: self.tool.clear_all()
        )

        self.actions["kernel.restart"] = QAction(QIcon(":/icons/reload.svg"),
                                                      trans("interpreter.menu.kernel.restart"))
        self.actions["kernel.restart"].triggered.connect(
            lambda: self.tool.restart_kernel()
        )

        # add actions
        self.menu["file"].addAction(self.actions["file.clear_output"])
        self.menu["file"].addAction(self.actions["file.clear_history"])
        self.menu["file"].addAction(self.actions["file.clear_all"])
        self.menu["kernel"].addAction(self.actions["kernel.restart"])
        return self.menu_bar

    def setup(self):
        """Setup interpreter dialog"""
        self.layout = self.widget.setup(all=True)
        self.layout.setMenuBar(self.setup_menu())  # add menu bar

        self.window.ui.dialog['interpreter'] = ToolDialog(self.window)
        self.window.ui.dialog['interpreter'].setLayout(self.layout)
        self.window.ui.dialog['interpreter'].setWindowTitle(trans("dialog.interpreter.title"))
        self.window.ui.dialog['interpreter'].resize(800, 500)


class ToolDialog(BaseDialog):
    def __init__(self, window=None, id="interpreter"):
        """
        Interpreter dialog

        :param window: main window
        :param id: logger id
        """
        super(ToolDialog, self).__init__(window, id)
        self.window = window

    def closeEvent(self, event):
        """
        Close event

        :param event: close event
        """
        self.cleanup()
        super(ToolDialog, self).closeEvent(event)

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key press event
        """
        if event.key() == Qt.Key_Escape:
            self.cleanup()
            self.close()  # close dialog when the Esc key is pressed.
        else:
            super(ToolDialog, self).keyPressEvent(event)

    def cleanup(self):
        """
        Cleanup on close
        """
        if self.window is None:
            return
        self.window.tools.get("interpreter").opened = False
        self.window.tools.get("interpreter").close()
        self.window.tools.get("interpreter").update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

from llama_index.core.readers.base import BaseReader
from llama_index.readers.web.sitemap.base import SitemapReader

from .base import BaseLoader


class Loader(BaseLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = "sitemap"
        self.name = "Sitemap"
        self.type = ["web"]
        self.instructions = [
            {
                "sitemap": {
                    "description": "read all web pages from sitemap.xml",
                    "args": {
                        "url": {
                            "type": "str",
                            "label": "URL",
                            "description": "URL to sitemap XML, e.g. https://example.com/sitemap.xml, all pages will be read",
                        },
                    },
                }
            }
        ]
        self.init_args = {
            "html_to_text": False,
            "limit": 10,
        }
        self.init_args_types = {
            "html_to_text": "bool",
            "limit": "int",
        }
        self.init_args_desc = {
            "html_to_text": "Whether to convert HTML to text",
            "limit": "Maximum number of concurrent requests",
        }

    def get(self) -> BaseReader:
        """
        Get reader instance

        :return: Data reader instance
        """
        args = self.get_args()
        return SitemapReader(**args)

    def prepare_args(self, **kwargs) -> dict:
        """
        Prepare arguments for reader

        :param kwargs: keyword arguments
        :return: args to pass to reader
        """
        args = {}
        args["sitemap_url"] = kwargs.get("url")  # list of links
        return args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 20:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_EXPERT,
)
from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event

from .config import Config

class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "experts"
        self.name = "Experts (inline)"
        self.description = "Enables inline experts in current mode."
        self.prefix = "Experts"
        self.type = [
            "expert",
        ]
        self.allowed_cmds = [
            "expert_call",
        ]
        self.order = 9998
        self.use_locale = True
        self.disallowed_modes = [MODE_AGENT, MODE_EXPERT]
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """
        Initialize options
        """
        self.config.from_defaults(self)

    def is_allowed(self) -> bool:
        """
        Check if plugin is allowed to run

        :return: True if plugin is allowed to run
        """
        return self.window.core.config.get("mode") not in self.disallowed_modes

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        if not self.is_allowed():
            return

        name = event.name
        data = event.data
        if name == Event.SYSTEM_PROMPT:
            if "is_expert" in data and data["is_expert"]:
                return  # disable if called inside expert
            data['value'] = self.on_system_prompt(data['value'])

    def on_system_prompt(self, prompt: str) -> str:
        """
        Event: SYSTEM_PROMPT

        :param prompt: prompt
        :return: updated prompt
        """
        if prompt is not None and prompt.strip() != "":
            prompt+= "\n\n"
        return prompt + self.window.core.experts.get_prompt()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.18 12:00:00                  #
# ================================================== #

import os
import cv2
import time

from PySide6.QtCore import QObject, Signal, QRunnable, Slot


class Camera:
    def __init__(self, window=None):
        """
        Camera core

        :param window: Window instance
        """
        self.window = window
        self.capture = None
        self.current = None

    def install(self):
        """Install provider data"""
        img_dir = self.window.core.config.get_user_dir('capture')
        if not os.path.exists(img_dir):
            os.makedirs(img_dir, exist_ok=True)


class CaptureSignals(QObject):
    finished = Signal()
    unfinished = Signal()
    destroyed = Signal()
    started = Signal()
    stopped = Signal()
    capture = Signal(object)
    error = Signal(object)


class CaptureWorker(QRunnable):
    def __init__(self, *args, **kwargs):
        super(CaptureWorker, self).__init__()
        self.signals = CaptureSignals()
        self.args = args
        self.kwargs = kwargs
        self.window = None
        self.initialized = False
        self.capture = None
        self.frame = None
        self.allow_finish = False

    def setup_camera(self):
        """Initialize camera"""
        try:
            # get params from global config
            self.capture = cv2.VideoCapture(self.window.core.config.get('vision.capture.idx'))
            if not self.capture.isOpened():
                self.allow_finish = False
                self.signals.unfinished.emit()
                return
            self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, self.window.core.config.get('vision.capture.width'))
            self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, self.window.core.config.get('vision.capture.height'))
        except Exception as e:
            self.window.core.debug.log(e)
            if self.signals is not None:
                self.signals.error.emit(e)
                self.signals.finished.emit(e)

    @Slot()
    def run(self):
        """Frame capture loop"""
        target_fps = 30
        fps_interval = 1.0 / target_fps
        self.allow_finish = True
        try:
            if not self.initialized:
                self.setup_camera()
                self.signals.started.emit()
                self.initialized = True
            last_frame_time = time.time()
            while True:
                if self.window.is_closing \
                        or self.capture is None \
                        or not self.capture.isOpened() \
                        or self.window.controller.camera.stop:
                    self.release()  # release camera
                    self.signals.stopped.emit()
                    break
                _, frame = self.capture.read()
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                now = time.time()
                if now - last_frame_time >= fps_interval:
                    self.signals.capture.emit(frame)
                    last_frame_time = now
        except Exception as e:
            self.window.core.debug.log(e)
            if self.signals is not None:
                self.signals.error.emit(e)

        # release camera
        self.release()
        if self.signals is not None:
            if self.allow_finish:
                self.signals.finished.emit()
            else:
                self.signals.unfinished.emit()

    def release(self):
        """Release camera"""
        if self.capture is not None and self.capture.isOpened():
            self.capture.release()
            self.capture = None
            self.frame = None
            self.initialized = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.08.25 04:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenu

from pygpt_net.utils import trans
from .base_combo import BaseCombo

class IndexCombo(BaseCombo):
    def __init__(self, window=None, parent_id: str = None, id: str = None, option: dict = None):
        """
        Index select combobox

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        """
        super(IndexCombo, self).__init__(window, parent_id, id, option)

    def on_combo_change(self, index):
        """
        On combo change

        :param index: combo index
        :return:
        """
        self.current_id = self.combo.itemData(index)
        self.window.controller.idx.select_by_id(self.current_id)

    def mousePressEvent(self, event):
        """
        Mouse press event

        :param event: event
        """
        if event.button() == Qt.RightButton:
            self.on_context_menu(self, event.pos())
        else:
            super().mousePressEvent(event)

    def on_context_menu(self, parent, pos):
        """
        Context menu event

        :param parent: parent widget
        :param pos: position
        """
        actions = {}
        actions['edit'] = QAction(QIcon(":/icons/edit.svg"), trans('action.edit'), self)
        actions['edit'].triggered.connect(
            lambda: self.action_edit()
        )

        txt = trans('idx.index_now') + ': ' + trans('settings.llama.extra.btn.idx_db_all')
        actions['idx_db_all'] = QAction(QIcon(":/icons/db.svg"), txt, self)
        actions['idx_db_all'].triggered.connect(
            lambda: self.action_idx_db_all()
        )

        txt = trans('idx.index_now') + ': ' + trans('settings.llama.extra.btn.idx_db_update')
        actions['idx_db_update'] = QAction(QIcon(":/icons/db.svg"), txt, self)
        actions['idx_db_update'].triggered.connect(
            lambda: self.action_idx_db_update()
        )

        txt = trans('idx.index_now') + ': ' + trans('settings.llama.extra.btn.idx_files_all')
        actions['idx_files_all'] = QAction(QIcon(":/icons/db.svg"), txt, self)
        actions['idx_files_all'].triggered.connect(
            lambda: self.action_idx_files_all()
        )

        actions['clear'] = QAction(QIcon(":/icons/close.svg"), trans('idx.btn.clear'), self)
        actions['clear'].triggered.connect(
            lambda: self.action_clear()
        )

        actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('idx.btn.truncate'), self)
        actions['delete'].triggered.connect(
            lambda: self.action_truncate()
        )

        menu = QMenu(self)
        menu.addAction(actions['edit'])
        menu.addAction(actions['idx_db_all'])
        menu.addAction(actions['idx_db_update'])
        menu.addAction(actions['idx_files_all'])
        menu.addAction(actions['clear'])
        menu.addAction(actions['delete'])
        menu.exec_(parent.mapToGlobal(pos))

    def action_idx_db_all(self):
        """Idx action handler"""
        if self.current_id is not None:
            self.window.controller.idx.indexer.index_ctx_from_ts(self.current_id, 0)

    def action_idx_db_update(self):
        """Idx action handler"""
        if self.current_id is not None:
            self.window.controller.idx.indexer.index_ctx_current(self.current_id)

    def action_idx_files_all(self):
        """Idx action handler"""
        if self.current_id is not None:
            self.window.controller.idx.indexer.index_all_files(self.current_id)

    def action_edit(self):
        """Edit action handler"""
        if self.current_id is not None:
            self.window.controller.settings.open_section('llama-index')

    def action_clear(self):
        """Clear idx action handler"""
        if self.current_id is not None:
            self.window.controller.idx.indexer.clear(self.current_id)

    def action_truncate(self):
        """Truncate idx action handler"""
        if self.current_id is not None:
            self.window.controller.idx.indexer.clear(self.current_id)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import datetime
from typing import Optional, List, Dict, Any

from packaging.version import Version

from pygpt_net.item.assistant import AssistantStoreItem, AssistantItem
from pygpt_net.provider.core.assistant_store.db_sqlite import DbSqliteProvider


class Store:
    def __init__(self, window=None):
        """
        Assistant vector store core

        :param window: Window instance
        """
        self.window = window
        self.provider = DbSqliteProvider(window)
        self.items = {}

    def install(self):
        """Install provider data"""
        self.provider.install()

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if data was patched
        """
        return self.provider.patch(app_version)

    def get(self, id: str) -> AssistantStoreItem:
        """
        Get store item by store_id

        :param id: store ID
        :return: store item
        """
        if id in self.items:
            return self.items[id]

    def get_ids(self) -> List[str]:
        """
        Return all loaded store IDs

        :return: store ids list
        """
        return list(self.items.keys())


    def get_all(self) -> Dict[str, AssistantStoreItem]:
        """
        Return all stores

        :return: all stores
        """
        return self.items

    def get_id_by_idx_all(self, idx: int) -> str:
        """
        Return store id by index

        :param idx: store idx
        :return: store id
        """
        return list(self.items.keys())[idx]


    def get_by_idx(self, idx: int) -> str:
        """
        Return store by index

        :param idx: store idx
        :return: model name
        """
        items = self.items
        return list(items.keys())[idx]


    def has(self, id: str) -> bool:
        """
        Check if store exists

        :param id: store id
        :return: True if store exists
        """
        return id in self.items

    def create(self) -> Optional[AssistantStoreItem]:
        """
        Create new store

        :return: store item
        """
        name = "New vector store"
        vector_store = self.window.core.gpt.store.create_store(name, 0)
        if vector_store is None:
            return None
        store = AssistantStoreItem()
        store.id = vector_store.id
        store.name = name
        store.is_thread = False
        store.record_id = self.provider.create(store)
        self.items[store.id] = store
        return store

    def update(self, store: AssistantStoreItem) -> Optional[AssistantStoreItem]:
        """
        Update store

        :param store: store
        :return: updated store or None if failed
        """
        vector_store = self.window.core.gpt.store.update_store(store.id, store.name, store.expire_days)
        if vector_store is None:
            return None
        self.items[store.id] = store
        self.provider.save(store)
        return store

    def get_status_data(self, id: str):
        """
        Get store status data

        :param id: store id
        :return: status data, store data
        """
        status = {}
        data = self.window.core.gpt.store.get_store(id)
        if data is not None:
            status = self.parse_status(data)
        return status, data

    def parse_status(self, store) -> Dict[str, Any]:
        """
        Parse store status

        :param store: store API object
        :return: status dict
        """
        status = {
            "status": store.status,
            "usage_bytes": store.usage_bytes,
            "expires_at": store.expires_at,
            "last_active_at": store.last_active_at,
        }
        if store.file_counts:
            status["file_counts"] = {
                "in_progress": store.file_counts.in_progress,
                "completed": store.file_counts.completed,
                "cancelled": store.file_counts.cancelled,
                "failed": store.file_counts.failed,
                "total": store.file_counts.total,
            }
        if store.expires_after:
            status["expires_after"] = {
                "anchor": store.expires_after.anchor,
                "days": store.expires_after.days,
            }
        return status

    def update_status(self, id: str):
        """
        Update store status

        :param id: store id
        """
        store = self.items[id]
        status, data = self.get_status_data(id)  # get from API
        tmp_name = data.name
        if tmp_name is None:
            tmp_name = ""
        store.name = tmp_name
        self.append_status(store, status)  # append to store
        self.update(store)  # save to db

    def append_status(
            self,
            store: AssistantStoreItem,
            status: Dict[str, Any]
    ):
        """
        Append status to store

        :param store: store
        :param status: status
        """
        now = datetime.datetime.now()
        ts = int(now.timestamp())
        status["__last_refresh__"] = now.strftime("%Y-%m-%d %H:%M:%S")
        store.status = status
        store.last_sync = ts
        if "status" in status:
            store.last_status = status["status"]
        if "usage_bytes" in status:
            store.usage_bytes = status["usage_bytes"]
        if "file_counts" in status:
            store.num_files = status["file_counts"]["total"]
        if "last_active_at" in status:
            store.last_active = int(status["last_active_at"])
        if "expires_after" in status:
            store.expire_days = int(status["expires_after"]["days"] or 0)

    def get_names(self) -> Dict[str, str]:
        """
        Get store name mapping id <> name

        :return: store names
        """
        names = {}
        for id in self.items:
            store = self.items[id]
            names[id] = store.name
        return names

    def delete(self, id: str) -> bool:
        """
        Delete store

        :param id: store id
        :return: True if store was deleted
        """
        if id in self.items:
            store = self.items[id]
            self.provider.delete_by_id(store.record_id)
            self.window.core.gpt.store.remove_store(id)
            del self.items[id]
            return True
        return False

    def import_items(self, items: Dict[str, AssistantStoreItem]):
        """
        Insert items

        :param items: items
        """
        self.items = items
        for item in items.values():
            item.record_id = self.provider.create(item)

    def clear(self):
        """Clear all stores"""
        self.truncate()

    def is_hidden(self, id: str) -> bool:
        """
        Check if store is hidden

        :param id: store id
        :return: True if store is hidden
        """
        if id in self.items:
            if (self.window.core.config.get("assistant.store.hide_threads")
                    and (self.items[id].name is None or self.items[id].name == "")):
                return True
        return False

    def truncate(self) -> bool:
        """
        Truncate all stores

        :return: True if truncated
        """
        self.provider.truncate()
        self.items = {}
        return True

    def load(self):
        """Load store"""
        self.items = self.provider.load_all()
        self.sort_items()

    def sort_items(self):
        """Sort items"""
        pass

    def save(self):
        """Save store"""
        self.provider.save_all(self.items)

    def get_version(self) -> str:
        """
        Get config version

        :return: version
        """
        return self.provider.get_version()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.20 06:00:00                  #
# ================================================== #

from PySide6.QtCore import QTimer
from PySide6.QtGui import QTextCursor, Qt
from PySide6.QtWebEngineCore import QWebEnginePage
from PySide6.QtWidgets import QTextEdit


class WebFinder:
    def __init__(self, window=None, textarea = None):
        """
        Finder core (in widget)

        :param window: Window instance
        :param textarea: QTextEdit / QTextBrowser (parent)
        """
        self.window = window
        self.textarea = textarea
        self.last_search = None
        self.matches = 0
        self.current_match_index = 0
        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.find_execute)
        self.delay = 100
        self.search_text = ""

    def assign(self, textarea):
        """
        Assign parent for the finder

        :param textarea: QTextEdit / QTextBrowser
        """
        self.textarea = textarea

    def text_changed(self):
        """On parent textarea text changed"""
        self.reset_highlights()

    def select(self):
        """Set current active"""
        current_search = self.get_search_string()
        if current_search != self.last_search:
            self.clear_search()
            self.search_text = current_search
            self.find_execute()
        else:
            if current_search is not None and len(current_search) > 0:
                self.clear_search()
                self.search_text = current_search
                self.find_execute()

    def clear_search(self):
        """Clear search state"""
        self.last_search = None
        self.matches = 0
        self.current_match_index = 0
        self.reset_highlights()

    def get_search_string(self) -> str:
        """
        Get current search string

        :return: search string
        """
        return self.window.ui.nodes['dialog.find.input'].text()

    def clear(self, restore: bool = False, to_end: bool = True):
        """
        Clear current parent highlights

        :param restore: Restore original document
        :param to_end: Move cursor to the end
        """
        self.clear_search()

    def parent(self):
        """
        Get current parent QTextEdit / QTextBrowser

        :return: QTextEdit / QTextBrowser
        """
        return self.textarea

    def prepare(self, clear: bool = True):
        """
        Prepare the finder for a new search

        :param clear: clear search state
        """
        if clear:
            self.clear_search()

    def find_execute(self):
        """Execute search"""
        text = self.search_text
        self.reset_highlights()
        if self.last_search is None or text != self.last_search:
            self.prepare()

        if text == self.last_search:
            self.matches = 0

        if text is None or len(text) == 0:
            self.last_search = None
            self.matches = 0
            self.reset_highlights()
            self.window.controller.finder.update_counter()
            return

        self.last_search = text
        self.handle_matches(text)
        if self.matches > 0:
            if self.current_match_index < 0:
                self.find_next()

        self.window.controller.finder.update_counter()

    def on_find_finished(self):
        self.window.controller.finder.update_counter()

    def find(self, text: str):
        """
        Find and highlight all occurrences of the text

        :param text: text to find
        """
        self.search_text = text
        self.timer.start(self.delay)

    def handle_matches(self, text: str):
        """
        Handle matches for the given text

        :param text: text to find
        """
        self.reset_highlights()
        parent = self.parent()
        parent.findText(text)

    def show_matches(self):
        """Show matches in the current parent"""
        return
        self.reset_highlights()  # Clear existing highlights
        parent = self.parent()

        cursor = QTextCursor(self.textarea.document())
        extra_selections = []
        current = self.current_match_index
        if current == -1:
            current = 0
        i = 0
        for start, length in self.matches:
            cursor.setPosition(start)
            cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, length)
            selection = QTextEdit.ExtraSelection()
            selection.format.setForeground(Qt.black)
            if i == current:
                selection.format.setBackground(Qt.green)
            else:
                selection.format.setBackground(Qt.yellow)
            selection.cursor = cursor
            extra_selections.append(selection)
            i += 1

        self.textarea.setExtraSelections(extra_selections)

    def reset_highlights(self):
        """Reset all highlights."""
        parent = self.parent()
        parent.findText("")

    def find_next(self):
        """Find the next occurrence relative to the current match index"""
       # self.current_match_index = (self.current_match_index + 1) % len(self.matches)
        parent = self.parent()
        parent.findText(self.search_text)

    def find_prev(self):
        """Find the previous occurrence relative to the current match index"""
        if not self.matches:
            pass
            #return

        #self.current_match_index = (self.current_match_index - 1) % len(self.matches)
        parent = self.parent()
        parent.findText(self.search_text, QWebEnginePage.FindBackward)
        #self.scroll_to(self.current_match_index)

    def scroll_to(self, match_index: int):
        """
        Scroll to the match given by match index

        :param match_index: current match index
        """
        return
        parent = self.parent()
        if match_index < 0 or match_index >= len(self.matches):
            return

        pos, _ = self.matches[match_index]
        cursor = QTextCursor(parent.document())
        cursor.setPosition(pos)
        parent.setTextCursor(cursor)
